"""Allows masking of PII (Personally Identifiable Information) in text documents."""

from __future__ import annotations

from collections.abc import Callable, Iterable, Sequence
from pathlib import Path
from typing import Any, override

import datasets
import dspy
import pydantic

from sieves.data.doc import Doc
from sieves.model_wrappers import ModelType
from sieves.model_wrappers.types import ModelSettings
from sieves.serialization import Config
from sieves.tasks.distillation.types import DistillationFramework
from sieves.tasks.predictive.core import PredictiveTask
from sieves.tasks.predictive.pii_masking.bridges import DSPyPIIMasking, PydanticPIIMasking
from sieves.tasks.predictive.schemas.pii_masking import (
    FewshotExample,
    TaskModel,
    TaskPromptSignature,
    TaskResult,
)

_TaskBridge = DSPyPIIMasking | PydanticPIIMasking


class PIIMasking(PredictiveTask[TaskPromptSignature, TaskResult, _TaskBridge]):
    """Task for masking PII (Personally Identifiable Information) in text documents."""

    def __init__(
        self,
        model: TaskModel,
        pii_types: Sequence[str] | dict[str, str] | None = None,
        task_id: str | None = None,
        include_meta: bool = True,
        batch_size: int = -1,
        overwrite: bool = False,
        prompt_instructions: str | None = None,
        fewshot_examples: Sequence[FewshotExample] = (),
        model_settings: ModelSettings = ModelSettings(),
        condition: Callable[[Doc], bool] | None = None,
    ) -> None:
        """Initialize PIIMasking task.

        :param model: Model to use.
        :param pii_types: Types of PII to mask. Supports three formats:
            - None (default): Uses all common PII types
            - List format: `["EMAIL", "PHONE", "SSN"]`
            - Dict format: `{"EMAIL": "Email addresses", "PHONE": "Phone numbers"}`
            The dict format allows you to provide descriptions that help the model better identify PII.
        :param task_id: Task ID.
        :param include_meta: Whether to include meta information generated by the task.
        :param batch_size: Batch size to use for inference. Use -1 to process all documents at once.
        :param overwrite: Whether to overwrite original document text with masked text.
        :param prompt_instructions: Custom prompt instructions. If None, default instructions are used.
        :param fewshot_examples: Few-shot examples.
        :param model_settings: Settings for structured generation.
        :param condition: Optional callable that determines whether to process each document.
        """
        if pii_types is not None:
            if isinstance(pii_types, dict):
                self._pii_types = list(pii_types.keys())
                self._pii_type_descriptions = pii_types
            else:
                self._pii_types = list(pii_types)
                self._pii_type_descriptions = {}
        else:
            self._pii_types = None
            self._pii_type_descriptions = {}

        self._pii_types_param = pii_types
        self._mask_placeholder = "[MASKED]"

        super().__init__(
            model=model,
            task_id=task_id,
            include_meta=include_meta,
            batch_size=batch_size,
            overwrite=overwrite,
            prompt_instructions=prompt_instructions,
            fewshot_examples=fewshot_examples,
            model_settings=model_settings,
            condition=condition,
        )

    @property
    @override
    def fewshot_example_type(self) -> type[FewshotExample]:
        """Return few-shot example type.

        :return: Few-shot example type.
        """
        return FewshotExample

    @property
    @override
    def prompt_signature(self) -> type[pydantic.BaseModel]:
        return TaskResult

    @property
    @override
    def metric(self) -> str:
        return "F1"

    @override
    def _compute_metrics(self, truths: list[Any], preds: list[Any], judge: dspy.LM | None = None) -> dict[str, float]:
        """Compute corpus-level metrics.

        :param truths: List of ground truths.
        :param preds: List of predictions.
        :param judge: Optional DSPy LM instance to use as judge for generative tasks.
        :return: Dictionary of metrics.
        """
        tp = 0
        fp = 0
        fn = 0

        for gold, pred in zip(truths, preds):
            # Extract entities.
            if gold is not None:
                assert isinstance(gold, TaskResult)
                true_entities = {(e.entity_type, e.text) for e in gold.pii_entities}
            else:
                true_entities = set()

            if pred is not None:
                assert isinstance(pred, TaskResult)
                pred_entities = {(e.entity_type, e.text) for e in pred.pii_entities}
            else:
                pred_entities = set()

            tp += len(true_entities & pred_entities)
            fp += len(pred_entities - true_entities)
            fn += len(true_entities - pred_entities)

        precision = tp / (tp + fp) if (tp + fp) > 0 else 0.0
        recall = tp / (tp + fn) if (tp + fn) > 0 else 0.0
        f1 = 2 * precision * recall / (precision + recall) if (precision + recall) > 0 else 0.0

        return {self.metric: f1}

    @override
    def _init_bridge(self, model_type: ModelType) -> _TaskBridge:
        bridge_types: dict[ModelType, type[_TaskBridge]] = {
            ModelType.dspy: DSPyPIIMasking,
            ModelType.langchain: PydanticPIIMasking,
            ModelType.outlines: PydanticPIIMasking,
        }

        try:
            return bridge_types[model_type](
                task_id=self._task_id,
                prompt_instructions=self._custom_prompt_instructions,
                mask_placeholder=self._mask_placeholder,
                pii_types=self._pii_types_param,
                overwrite=self._overwrite,
                model_settings=self._model_settings,
                prompt_signature=self.prompt_signature,
                model_type=model_type,
                fewshot_examples=self._fewshot_examples,
            )
        except KeyError as err:
            raise KeyError(f"Model type {model_type} is not supported by {self.__class__.__name__}.") from err

    @staticmethod
    @override
    def supports() -> set[ModelType]:
        return {
            ModelType.dspy,
            ModelType.langchain,
            ModelType.outlines,
        }

    @property
    @override
    def _state(self) -> dict[str, Any]:
        # Store pii_types as dict if descriptions exist, else as original value
        pii_types_state = self._pii_type_descriptions if self._pii_type_descriptions else self._pii_types
        return {
            **super()._state,
            "pii_types": pii_types_state,
        }

    @override
    def to_hf_dataset(self, docs: Iterable[Doc], threshold: float | None = None) -> datasets.Dataset:
        # Define metadata.
        features = datasets.Features(
            {
                "text": datasets.Value("string"),
                "masked_text": datasets.Value("string"),
            }
        )
        info = datasets.DatasetInfo(
            description=f"PII masking dataset. Generated with sieves v{Config.get_version()}.",
            features=features,
        )

        # Fetch data used for generating dataset.
        try:
            data = [(doc.text, doc.results[self._task_id].masked_text) for doc in docs]
        except KeyError as err:
            raise KeyError(f"Not all documents have results for this task with ID {self._task_id}") from err

        def generate_data() -> Iterable[dict[str, Any]]:
            """Yield results as dicts.

            :return: Results as dicts.
            """
            for text, masked_text in data:
                yield {"text": text, "masked_text": masked_text}

        # Create dataset.
        return datasets.Dataset.from_generator(generate_data, features=features, info=info)

    @override
    def distill(
        self,
        base_model_id: str,
        framework: DistillationFramework,
        data: datasets.Dataset | Sequence[Doc],
        output_path: Path | str,
        val_frac: float,
        init_kwargs: dict[str, Any] | None = None,
        train_kwargs: dict[str, Any] | None = None,
        seed: int | None = None,
    ) -> None:
        raise NotImplementedError

    @override
    def _evaluate_dspy_example(self, truth: dspy.Example, pred: dspy.Prediction, trace: Any, model: dspy.LM) -> float:
        # Compute entity detection F1 score based on (entity_type, text) pairs
        true_entities = {(e["entity_type"], e["text"]) for e in truth["pii_entities"]}
        pred_entities = {(e["entity_type"], e["text"]) for e in pred.get("pii_entities", [])}

        if not true_entities:
            return 1.0 if not pred_entities else 0.0

        precision = len(true_entities & pred_entities) / len(pred_entities) if pred_entities else 0
        recall = len(true_entities & pred_entities) / len(true_entities)
        return 2 * precision * recall / (precision + recall) if (precision + recall) > 0 else 0



"""Optimizer implementation."""

import random
from collections.abc import Callable
from typing import Any, Self

import dspy

from sieves.serialization import Attribute, Config

EvalMetric = Callable[[dspy.Example, dspy.Prediction, Any], float]


class Optimizer:
    """Config for task optimization with DSPy.

    Uses MIPROv2 to optimize instructions and few-shot examples.
    """

    def __init__(
        self,
        model: dspy.LM | dspy.BaseLM,
        val_frac: float,
        seed: int | None = None,
        shuffle: bool = True,
        dspy_init_kwargs: dict[str, Any] | None = None,
        dspy_compile_kwargs: dict[str, Any] | None = None,
    ):
        """Initialize optimizer.

        :param model: Fully initialized DSPy model to use for optimization. Doesn't have to be the same as the model
            used to run the task, but more similar is better. With a lot of data you might want to pick a faster/cheaper
            model.
        :param val_frac: Fraction of examples to use for validation. Everything else is used for optimization.
        :param seed: Random seed for data splitting.
        :param shuffle: Whether to shuffle the data.
        :param dspy_init_kwargs: Optional keyword arguments to pass to DSPy optimizer at init time.
        :param dspy_compile_kwargs: Optional keyword arguments to pass to DSPy optimizer at compile time.
        """
        self._model = model
        self._val_frac = val_frac
        self._seed = seed
        self._shuffle = shuffle
        self._init_kwargs = dspy_init_kwargs or {}
        self._compile_kwargs = {"requires_permission_to_run": False} | (dspy_compile_kwargs or {})

    def __call__(
        self,
        signature: type[dspy.Signature] | type[dspy.Module],
        data: list[dspy.Example],
        evaluate: EvalMetric,
        verbose: bool = False,
    ) -> tuple[str, list[dspy.Example]]:
        """Optimize prompt and few-shot examples w.r.t. given signature and dataset.

        :param signature: Task to optimize.
        :param data: Dataset to use for optimization.
        :param evaluate: Evaluation metric to use for optimization.
        :param verbose: Whether to log DSPy output.
        :return: Best combination of (1) prompt and (2) fewshot-examples.
        """
        predictor = dspy.Predict(signature)
        teleprompter = dspy.MIPROv2(metric=evaluate, **(self._init_kwargs or {}), verbose=False)
        trainset, devset = self._split_data(data, self._val_frac, self._seed, self._shuffle)

        optimized_predictor: dspy.Predict = teleprompter.compile(
            predictor, trainset=trainset, valset=devset, **(self._compile_kwargs or {})
        )

        return optimized_predictor.signature.instructions, optimized_predictor.demos

    @property
    def model(self) -> dspy.LM:
        """Return model used for optimization.

        :return dspy.LM: Model used for optimization.
        """
        assert isinstance(self._model, dspy.LM)
        return self._model

    @property
    def _state(self) -> dict[str, Any]:
        """Return attributes to serialize.

        :return: Dict of attributes to serialize.
        """
        return {
            "model": self._model,
            "val_frac": self._val_frac,
            "seed": self._seed,
            "shuffle": self._shuffle,
            "init_kwargs": self._init_kwargs,
            "compile_kwargs": self._compile_kwargs,
        }

    def serialize(self) -> Config:
        """Serialize task.

        :return: Config instance.
        """
        return Config.create(self.__class__, {k: Attribute(value=v) for k, v in self._state.items()})

    @classmethod
    def deserialize(cls, config: Config, **kwargs: dict[str, Any]) -> Self:
        """Generate Optimizer instance from config.

        :param config: Config to generate instance from.
        :param kwargs: Values to inject into loaded config.
        :return: Deserialized Optimizer instance.
        """
        return cls(**config.to_init_dict(cls, **kwargs))

    @staticmethod
    def _split_data(
        data: list[dspy.Example], val_frac: float, seed: int | None, shuffle: bool
    ) -> tuple[list[dspy.Example], list[dspy.Example]]:
        """Split data into train and validation sets.

        :param data: Dataset to split.
        :param val_frac: Fraction of data to use for validation.
        :param seed: Random seed for shuffling.
        :param shuffle: Whether to shuffle the data before splitting.
        :return: Tuple of (trainset, valset).
        """
        dataset = data.copy()
        if shuffle:
            rng = random.Random(seed)
            rng.shuffle(dataset)

        val_size = int(len(dataset) * val_frac)
        trainset = dataset[val_size:]
        valset = dataset[:val_size]

        return trainset, valset



"""Predictive tasks."""

from sieves.tasks.predictive.classification import Classification
from sieves.tasks.predictive.core import PredictiveTask
from sieves.tasks.predictive.information_extraction import InformationExtraction
from sieves.tasks.predictive.ner import NER
from sieves.tasks.predictive.pii_masking import PIIMasking
from sieves.tasks.predictive.question_answering import QuestionAnswering
from sieves.tasks.predictive.relation_extraction import RelationExtraction
from sieves.tasks.predictive.sentiment_analysis import SentimentAnalysis
from sieves.tasks.predictive.summarization import Summarization
from sieves.tasks.predictive.translation import Translation

__all__ = [
    "Classification",
    "InformationExtraction",
    "SentimentAnalysis",
    "Summarization",
    "Translation",
    "NER",
    "PIIMasking",
    "PredictiveTask",
    "QuestionAnswering",
    "RelationExtraction",
]



"""Model type enum and utilities."""

from __future__ import annotations

import enum

from sieves.model_wrappers.core import (
    ModelWrapper,
    ModelWrapperInferenceMode,
    ModelWrapperModel,
    ModelWrapperPromptSignature,
    ModelWrapperResult,
)
from sieves.model_wrappers.dspy_ import DSPy
from sieves.model_wrappers.gliner_ import GliNER
from sieves.model_wrappers.huggingface_ import HuggingFace
from sieves.model_wrappers.langchain_ import LangChain
from sieves.model_wrappers.outlines_ import Outlines


class ModelType(enum.Enum):
    """Available model types."""

    dspy = DSPy
    gliner = GliNER
    huggingface = HuggingFace
    langchain = LangChain
    outlines = Outlines

    @classmethod
    def all(cls) -> tuple[ModelType, ...]:
        """Return all available model types.

        :return tuple[ModelType, ...]: All available model types.
        """
        return tuple(ModelType)

    @classmethod
    def get_model_type(
        cls,
        model_wrapper: ModelWrapper[
            ModelWrapperPromptSignature, ModelWrapperResult, ModelWrapperModel, ModelWrapperInferenceMode
        ],
    ) -> ModelType:
        """Return model type for specified model wrapper.

        :param model_wrapper: ModelWrapper to get type for.
        :return ModelType: Model type for self._model_wrapper.
        :raises ValueError: if model wrapper class not found in ModelType.
        """
        for mt in ModelType:
            if isinstance(model_wrapper, mt.value):
                return mt
        raise ValueError(f"ModelWrapper class {model_wrapper.__class__.__name__} not found in ModelType.")



"""Named‑Entity Recognition (NER) predictive task."""

from __future__ import annotations

from collections.abc import Callable, Iterable, Sequence
from pathlib import Path
from typing import Any, Literal, override

import datasets
import dspy
import pydantic

from sieves.data import Doc
from sieves.model_wrappers import ModelType
from sieves.model_wrappers.types import ModelSettings
from sieves.serialization import Config
from sieves.tasks.distillation.types import DistillationFramework
from sieves.tasks.predictive.core import PredictiveTask
from sieves.tasks.predictive.gliner_bridge import GliNERBridge
from sieves.tasks.predictive.ner.bridges import DSPyNER, PydanticNER
from sieves.tasks.predictive.schemas.ner import (
    FewshotExample,
    TaskModel,
    TaskPromptSignature,
    TaskResult,
)

_TaskBridge = DSPyNER | GliNERBridge | PydanticNER


class NER(PredictiveTask[TaskPromptSignature, TaskResult, _TaskBridge]):
    """Extract named entities from text using various model wrappers."""

    def __init__(
        self,
        entities: Sequence[str] | dict[str, str] | None = None,
        model: TaskModel | None = None,
        task_id: str | None = None,
        include_meta: bool = True,
        batch_size: int = -1,
        prompt_instructions: str | None = None,
        fewshot_examples: Sequence[FewshotExample] = (),
        model_settings: ModelSettings = ModelSettings(),
        condition: Callable[[Doc], bool] | None = None,
    ) -> None:
        """Initialize NER task.

        :param entities: Entity types to extract. Supports two formats:
            - List format: `["PERSON", "LOCATION", "ORGANIZATION"]`
            - Dict format: `{"PERSON": "Names of people", "LOCATION": "Geographic locations"}`
            The dict format allows you to provide descriptions that help the model better identify entities.
            When using GliNER, descriptions are passed directly to the model for improved recognition.
            If None, defaults to `["PERSON", "LOCATION", "ORGANIZATION"]`.
        :param model: Model to use.
        :param task_id: Task ID.
        :param include_meta: Whether to include meta information generated by the task.
        :param batch_size: Batch size to use for inference. Use -1 to process all documents at once.
        :param prompt_instructions: Custom prompt instructions. If None, default instructions are used.
        :param fewshot_examples: Few-shot examples.
        :param model_settings: Settings for structured generation.
        :param condition: Optional callable that determines whether to process each document.
        """
        if entities is None:
            entities = ["PERSON", "LOCATION", "ORGANIZATION"]

        if isinstance(entities, dict):
            self._entities = list(entities.keys())
            self._entity_descriptions = entities
        else:
            self._entities = list(entities)
            self._entity_descriptions = {}

        self._entities_param = entities

        super().__init__(
            model=model,
            task_id=task_id,
            include_meta=include_meta,
            batch_size=batch_size,
            overwrite=False,
            prompt_instructions=prompt_instructions,
            fewshot_examples=fewshot_examples,
            model_settings=model_settings,
            condition=condition,
        )

    @property
    @override
    def fewshot_example_type(self) -> type[FewshotExample]:
        return FewshotExample

    @property
    @override
    def prompt_signature(self) -> type[pydantic.BaseModel]:
        """Return the unified Pydantic prompt signature for this task.

        :return: Unified Pydantic prompt signature.
        """
        # Create a dynamic entity model with Literal for the entity types.
        EntityTypes = Literal[*(tuple(self._entities))] if self._entities else str  # type: ignore[valid-type]

        DynamicEntity = pydantic.create_model(
            "NEREntity",
            __doc__="Extracted named entity with its context and type.",
            text=(str, pydantic.Field(description="The specific text segment identified as an entity.")),
            context=(str, pydantic.Field(description="The surrounding text providing context for the entity.")),
            entity_type=(
                EntityTypes,
                pydantic.Field(description="The category or type of the entity (e.g., PERSON, ORGANIZATION)."),
            ),
            score=(
                float | None,
                pydantic.Field(
                    default=None,
                    description="Provide a confidence score for the entity identification, between 0 and 1.",
                ),
            ),
            __base__=pydantic.BaseModel,
        )

        return pydantic.create_model(
            "NEROutput",
            __doc__="Result of named-entity recognition. Contains a list of extracted entities.",
            entities=(
                list[DynamicEntity],  # type: ignore[valid-type]
                pydantic.Field(..., description="List of extracted named entities."),
            ),
        )

    @property
    @override
    def metric(self) -> str:
        return "F1"

    @override
    def _compute_metrics(self, truths: list[Any], preds: list[Any], judge: dspy.LM | None = None) -> dict[str, float]:
        """Compute corpus-level metrics.

        :param truths: List of ground truths.
        :param preds: List of predictions.
        :param judge: Optional DSPy LM instance to use as judge for generative tasks.
        :return: Dictionary of metrics.
        """
        tp = 0
        fp = 0
        fn = 0

        for gold, pred in zip(truths, preds):
            if gold is None:
                true_entities = set()
            else:
                # Convert to DSPy format first to reuse logic or do it directly.
                assert isinstance(gold, TaskResult)
                true_entities = {(e.text, e.entity_type) for e in gold.entities}

            if pred is None:
                pred_entities = set()
            else:
                if hasattr(pred, "entities"):
                    pred_entities = {(e.text, e.entity_type) for e in pred.entities}
                elif isinstance(pred, dict) and "entities" in pred:
                    pred_entities = {(e["text"], e["entity_type"]) for e in pred.get("entities", [])}
                else:
                    pred_entities = set()

            tp += len(true_entities & pred_entities)
            fp += len(pred_entities - true_entities)
            fn += len(true_entities - pred_entities)

        precision = tp / (tp + fp) if (tp + fp) > 0 else 0.0
        recall = tp / (tp + fn) if (tp + fn) > 0 else 0.0
        f1 = 2 * precision * recall / (precision + recall) if (precision + recall) > 0 else 0.0

        return {self.metric: f1}

    @override
    def _init_bridge(self, model_type: ModelType) -> _TaskBridge:
        if model_type == ModelType.gliner:
            from sieves.model_wrappers import gliner_

            return GliNERBridge(
                task_id=self._task_id,
                prompt_instructions=self._custom_prompt_instructions,
                prompt_signature=self.prompt_signature,
                model_settings=self._model_settings,
                inference_mode=gliner_.InferenceMode.entities,
            )

        bridge_types = {
            ModelType.langchain: PydanticNER,
            ModelType.outlines: PydanticNER,
            ModelType.dspy: DSPyNER,
        }

        try:
            bridge_class = bridge_types[model_type]
            result = bridge_class(
                task_id=self._task_id,
                prompt_instructions=self._custom_prompt_instructions,
                entities=self._entities_param,
                model_settings=self._model_settings,
                prompt_signature=self.prompt_signature,
                model_type=model_type,
                fewshot_examples=self._fewshot_examples,
            )
            return result  # type: ignore[return-value]
        except KeyError as err:
            raise KeyError(f"Model type {model_type} is not supported by {self.__class__.__name__}.") from err

    @staticmethod
    @override
    def supports() -> set[ModelType]:
        return {
            ModelType.langchain,
            ModelType.dspy,
            ModelType.outlines,
            ModelType.gliner,
        }

    @override
    def _validate_fewshot_examples(self) -> None:
        for fs_example in self._fewshot_examples or []:
            assert isinstance(fs_example, FewshotExample)
            for entity in fs_example.entities:
                if entity.entity_type not in self._entities:
                    raise ValueError(f"Entity {entity.entity_type} not in {self._entities}.")

    @override
    @property
    def _state(self) -> dict[str, Any]:
        # Store entities as dict if descriptions exist, else as list
        entities_state = self._entity_descriptions if self._entity_descriptions else self._entities
        return {
            **super()._state,
            "entities": entities_state,
        }

    @override
    def to_hf_dataset(self, docs: Iterable[Doc], threshold: float | None = None) -> datasets.Dataset:
        # Define metadata and features for the dataset
        features = datasets.Features(
            {
                "text": datasets.Value("string"),
                "entities": datasets.Sequence(
                    datasets.Features(
                        {
                            "text": datasets.Value("string"),
                            "start": datasets.Value("int32"),
                            "end": datasets.Value("int32"),
                            "entity_type": datasets.Value("string"),
                            "score": datasets.Value("float32"),
                        }
                    )
                ),
            }
        )

        info = datasets.DatasetInfo(
            description=f"Named Entity Recognition dataset with entity types {self._entities}. Generated with sieves "
            f"v{Config.get_version()}.",
            features=features,
        )

        # Fetch data used for generating dataset
        try:
            data: list[tuple[str, list[dict[str, Any]]]] = []
            for doc in docs:
                if self._task_id not in doc.results:
                    raise KeyError(f"Document does not have results for task ID {self._task_id}")

                # Get the entities from the document results
                result = doc.results[self._task_id].entities
                entities: list[dict[str, Any]] = []

                # List format (could be list of dictionaries or other entities)
                for entity in result:
                    assert hasattr(entity, "text")
                    assert hasattr(entity, "start")
                    assert hasattr(entity, "end")
                    assert hasattr(entity, "entity_type")

                    entities.append(
                        {
                            "text": entity.text,
                            "start": entity.start,
                            "end": entity.end,
                            "entity_type": entity.entity_type,
                            "score": getattr(entity, "score", None),
                        }
                    )

                data.append((doc.text or "", entities))

        except KeyError as err:
            raise KeyError(f"Not all documents have results for this task with ID {self._task_id}") from err

        def generate_data() -> Iterable[dict[str, Any]]:
            """Yield results as dicts.

            :return: Results as dicts.
            """
            for text, entities in data:
                yield {"text": text, "entities": entities}

        # Create dataset
        return datasets.Dataset.from_generator(generate_data, features=features, info=info)

    @override
    def distill(
        self,
        base_model_id: str,
        framework: DistillationFramework,
        data: datasets.Dataset | Sequence[Doc],
        output_path: Path | str,
        val_frac: float,
        init_kwargs: dict[str, Any] | None = None,
        train_kwargs: dict[str, Any] | None = None,
        seed: int | None = None,
    ) -> None:
        raise NotImplementedError

    @override
    def _evaluate_dspy_example(self, truth: dspy.Example, pred: dspy.Prediction, trace: Any, model: dspy.LM) -> float:
        # Compute entity-level F1 score based on (text, entity_type) pairs
        true_entities = {(e["text"], e["entity_type"]) for e in truth["entities"]}
        pred_entities = {(e["text"], e["entity_type"]) for e in pred.get("entities", [])}

        if not true_entities:
            return 1.0 if not pred_entities else 0.0

        precision = len(true_entities & pred_entities) / len(pred_entities) if pred_entities else 0
        recall = len(true_entities & pred_entities) / len(true_entities)
        return 2 * precision * recall / (precision + recall) if (precision + recall) > 0 else 0



"""Allows chunking of documents into segments."""

import itertools
import sys
from collections.abc import Callable, Iterable
from typing import Any

import chonkie

from sieves.data.doc import Doc
from sieves.tasks.core import Task


class Chonkie(Task):
    """Chunker wrapping the chonkie library."""

    def __init__(
        self,
        chunker: chonkie.BaseChunker,
        task_id: str | None = None,
        include_meta: bool = False,
        batch_size: int = -1,
        condition: Callable[[Doc], bool] | None = None,
    ):
        """Initialize chunker.

        :param task_id: Task ID.
        :param include_meta: Whether to include meta information generated by the task.
        :param batch_size: Batch size to use for processing. Use -1 to process all documents at once.
        :param condition: Optional callable that determines whether to process each document.
        """
        super().__init__(task_id=task_id, include_meta=include_meta, batch_size=batch_size, condition=condition)
        self._chunker = chunker

    def _call(self, docs: Iterable[Doc]) -> Iterable[Doc]:
        """Split documents into chunks.

        :param docs: Documents to split.
        :return: Split documents.
        """
        batch_size = self._batch_size if self._batch_size > 0 else sys.maxsize
        while docs_batch := [doc for doc in itertools.islice(docs, batch_size)]:
            if len(docs_batch) == 0:
                break

            # Signature varies in between versions, try both.
            try:
                chunks = self._chunker.chunk_batch([doc.text for doc in docs_batch], show_progress=False)
            except TypeError:
                chunks = self._chunker.chunk_batch([doc.text for doc in docs_batch], show_progress_bar=False)  # type: ignore[unknown-argument]
            assert len(chunks) == len(docs_batch)

            for doc, doc_chunks in zip(docs_batch, chunks):
                if self._include_meta:
                    doc.meta |= {self.id: {doc_chunks}}
                doc.chunks = [chunk.text for chunk in doc_chunks]

                yield doc

    @property
    def _state(self) -> dict[str, Any]:
        return {
            **super()._state,
            "chunker": self._chunker,
        }



"""Common types."""

from sieves.model_wrappers import (
    dspy_,
    gliner_,
    huggingface_,
    langchain_,
    outlines_,
)

Model = dspy_.Model | gliner_.Model | huggingface_.Model | langchain_.Model | outlines_.Model



# mypy: ignore-errors
import traceback
from typing import Literal

import pydantic
import pytest
from flaky import flaky

from sieves import Doc, Pipeline
from sieves.model_wrappers import ModelType, ModelSettings
from sieves.serialization import Config
from sieves.tasks import Classification
from sieves.tasks.predictive import classification
from sieves.tests.conftest import Runtime


def _run(
    runtime: Runtime, docs: list[Doc],
    fewshot: bool,
    mode: Literal['multi', 'single'] = 'multi',
    test_hf_conversion: bool = False
) -> None:
    """Tests whether the classification task works as expected."""
    if mode == 'multi':
        fewshot_examples = [
            classification.FewshotExampleMultiLabel(
                text="On the properties of hydrogen atoms and red dwarfs.",
                score_per_label={"science": 1.0, "politics": 0.0},
            ),
            classification.FewshotExampleMultiLabel(
                text="A parliament is elected by casting votes.",
                score_per_label={"science": 0, "politics": 1.0},
            ),
        ]
    else:
        fewshot_examples = [
            classification.FewshotExampleSingleLabel(
                text="On the properties of hydrogen atoms and red dwarfs.",
                label="science",
                score=1.0,
            ),
            classification.FewshotExampleSingleLabel(
                text="A parliament is elected by casting votes.",
                label="politics",
                score=1.0,
            ),
        ]

    fewshot_args = {"fewshot_examples": fewshot_examples} if fewshot else {}
    labels = {
        "science": "Topics related to scientific disciplines and research",
        "politics": "Topics related to government, elections, and political systems",
    }

    task = classification.Classification(
        task_id="classifier",
        labels=labels,
        model=runtime.model,
        model_settings=runtime.model_settings,
        batch_size=runtime.batch_size,
        mode=mode,
        **fewshot_args,
    )
    pipe = Pipeline(task)
    docs = list(pipe(docs))

    assert len(docs) == 2
    for doc in docs:
        assert doc.text
        assert doc
        assert "classifier" in doc.results

        # Verify unified result types.
        if mode == 'multi':
            assert isinstance(doc.results["classifier"], classification.ResultMultiLabel)
        else:
            assert isinstance(doc.results["classifier"], classification.ResultSingleLabel)

        assert "classifier" in doc.meta
        assert "raw" in doc.meta["classifier"]
        assert "usage" in doc.meta["classifier"]
        assert "usage" in doc.meta

        print(f"Output: {doc.results['classifier']}")
        print(f"Raw output: {doc.meta['classifier']['raw']}")
        print(f"Usage: {doc.meta['classifier']['usage']}")
        print(f"Total Usage: {doc.meta['usage']}")

    if test_hf_conversion:
        _to_hf_dataset(task, docs, mode)

@pytest.mark.parametrize("batch_runtime", Classification.supports(), indirect=["batch_runtime"])
@pytest.mark.parametrize("fewshot", [True, False])
@pytest.mark.parametrize("mode", ['multi', 'single'])
def test_run(classification_docs, batch_runtime, fewshot, mode):
    try:
        _run(batch_runtime, classification_docs, fewshot, mode, test_hf_conversion=fewshot is True)
    except RuntimeError as err:
        # Outlines via OpenRouter/OpenAI API cannot deal with `Literal`s, hence classification may fail.
        # This is tolerable, but we should keep an eye on this and remove this fallback once possible.
        tbe = traceback.TracebackException.from_exception(err)
        stack_frames = traceback.extract_stack()
        tbe.stack.extend(stack_frames)
        if "The `openai` library does not support batch inference." not in ''.join(tbe.format()):
            raise err

@flaky(max_runs=3, min_passes=1)
@pytest.mark.parametrize("runtime", Classification.supports(), indirect=["runtime"])
@pytest.mark.parametrize("fewshot", [True, False])
def test_run_nonbatched(classification_docs, runtime, fewshot):
    _run(runtime, classification_docs, fewshot, test_hf_conversion=False)


def _to_hf_dataset(task: Classification, docs: list[Doc], mode: Literal['single', 'multi']) -> None:
    """Tests whether conversion to HF dataset works as expected.

    :param task: Classification task instance.
    :param docs: List of documents to convert.
    :param mode: Task mode.
    """
    dataset = task.to_hf_dataset(docs)
    assert all([key in dataset.features for key in ("text", "labels")])
    assert len(dataset) == 2
    dataset_records = list(dataset)
    for rec in dataset_records:
        if mode == 'multi':
            assert isinstance(rec["labels"], list)
            assert all(
                [isinstance(v, int) for v in rec["labels"]]
            ), "Labels should be integers for multi-label classification"
            for v in rec["labels"]:
                assert isinstance(v, int)
        else:
            assert isinstance(rec["labels"], int)
        assert isinstance(rec["text"], str)

    with pytest.raises(KeyError):
        task.to_hf_dataset([Doc(text="This is a dummy text.")])


@pytest.mark.parametrize("batch_runtime", [ModelType.huggingface], indirect=["batch_runtime"])
def test_serialization(classification_docs, batch_runtime) -> None:
    labels = {
        "science": "Topics related to scientific disciplines and research",
        "politics": "Topics related to government, elections, and political systems",
    }

    pipe = Pipeline(
        classification.Classification(
            task_id="classifier",
            labels=labels,
            model=batch_runtime.model,
            model_settings=batch_runtime.model_settings,
            batch_size=batch_runtime.batch_size,
        )
    )

    config = pipe.serialize()
    assert config.model_dump() == {'cls_name': 'sieves.pipeline.core.Pipeline',
                                   'tasks': {'is_placeholder': False,
                                             'value': [{
                                                           'cls_name': 'sieves.tasks.predictive.classification.core.Classification',
                                                           'fewshot_examples': {'is_placeholder': False,
                                                                                'value': ()},
                                                           'batch_size': {'is_placeholder': False, "value": -1},
                                                           'model_settings': {'is_placeholder': False,
                                                                                   'value': {
                                                                                       'config_kwargs': None,
                                                                                       'inference_kwargs': None,
                                                                                       'init_kwargs': None,
                                                                                       'strict': True,
                                                                                       'inference_mode': None}},
                                                           'include_meta': {'is_placeholder': False, 'value': True},
                                                           'labels': {'is_placeholder': False,
                                                                      'value': {'politics': 'Topics '
                                                                                                        'related to '
                                                                                                        'government, '
                                                                                                        'elections, '
                                                                                                        'and '
                                                                                                        'political '
                                                                                                        'systems',
                                                                                            'science': 'Topics '
                                                                                                       'related to '
                                                                                                       'scientific '
                                                                                                       'disciplines '
                                                                                                       'and '
                                                                                                       'research'}},
                                                           'model': {'is_placeholder': True,
                                                                     'value': 'transformers.pipelines.zero_shot_classification.ZeroShotClassificationPipeline'},
                                                           'prompt_instructions': {'is_placeholder': False,
                                                                                   'value': None},
                                                           'task_id': {'is_placeholder': False,
                                                                       'value': 'classifier'},
                                                           'condition': {'is_placeholder': False, 'value': None},
                                                           'version': Config.get_version()}]},
                                   'use_cache': {'is_placeholder': False, 'value': True},
                                   'version': Config.get_version()}

    Pipeline.deserialize(config=config, tasks_kwargs=[{"model": batch_runtime.model}])


@pytest.mark.parametrize("batch_runtime", [ModelType.huggingface], indirect=["batch_runtime"])
def test_labels_validation(batch_runtime) -> None:
    """Test that labels parameter accepts both list and dict formats."""
    # Valid case - list format (no descriptions)
    classification.Classification(
        labels=["science", "politics"],
        model=batch_runtime.model,
        model_settings=batch_runtime.model_settings,
        batch_size=batch_runtime.batch_size,
    )

    # Valid case - dict format with all labels having descriptions
    labels_with_descriptions = {"science": "Science related", "politics": "Politics related"}
    classification.Classification(
        labels=labels_with_descriptions,
        model=batch_runtime.model,
        model_settings=batch_runtime.model_settings,
        batch_size=batch_runtime.batch_size,
    )

    # Valid case - dict format with some labels having descriptions (empty strings)
    partial_descriptions = {"science": "Science related", "politics": ""}
    classification.Classification(
        labels=partial_descriptions,
        model=batch_runtime.model,
        model_settings=batch_runtime.model_settings,
        batch_size=batch_runtime.batch_size,
    )


def test_fewshot_example_singlelabel_score() -> None:
    """Test that the score of a fewshot example is correctly validated."""
    classification.FewshotExampleSingleLabel(
        text="...",
        label="science",
        score=1.0,
    )

    with pytest.raises(ValueError):
        classification.FewshotExampleSingleLabel(
            text="...",
            label="science",
            score=2.0,
        )

    with pytest.raises(ValueError):
        classification.FewshotExampleSingleLabel(
            text="...",
            label="science",
            score=-2.0,
        )


def test_result_to_scores() -> None:
    """Test that the result to scores method works as expected."""
    # 1) ResultMultiLabel
    res = classification.ResultMultiLabel(label_scores=[("science", 0.9), ("politics", 0.1)])
    scores = classification.Classification._result_to_scores(res)
    assert scores == {"science": 0.9, "politics": 0.1}

    # 2) ResultSingleLabel
    res = classification.ResultSingleLabel(label="science", score=0.8)
    scores = classification.Classification._result_to_scores(res)
    assert scores == {"science": 0.8}

    # 3) Unsupported types raise TypeError
    with pytest.raises(TypeError):
        classification.Classification._result_to_scores([("science", 0.9)])

    with pytest.raises(TypeError):
        classification.Classification._result_to_scores({"science": 0.9})

    with pytest.raises(TypeError):
        classification.Classification._result_to_scores("science")


@pytest.mark.parametrize("batch_runtime", Classification.supports(), indirect=["batch_runtime"])
def test_inference_mode_override(batch_runtime) -> None:
    """Test that inference_mode parameter overrides the default value."""
    dummy = "dummy_inference_mode"

    task = classification.Classification(
        task_id="classifier",
        labels=["science", "politics"],
        model=batch_runtime.model,
        model_settings=ModelSettings(inference_mode=dummy),
        batch_size=batch_runtime.batch_size,
    )

    assert task._bridge.inference_mode == dummy


@pytest.mark.parametrize("batch_runtime", [ModelType.huggingface], indirect=["batch_runtime"])
def test_evaluation(batch_runtime) -> None:
    """Test evaluation for classification without running pipeline."""
    labels = ["science", "politics"]
    task = Classification(labels=labels, model=batch_runtime.model, task_id="clf", mode="single")

    # 1. Full overlap
    doc_full = Doc(text="Text about science.")
    doc_full.results["clf"] = classification.ResultSingleLabel(label="science", score=0.9)
    doc_full.gold["clf"] = "science"

    report_full = task.evaluate([doc_full])
    assert report_full.metrics[task.metric] == 1.0

    # 2. No overlap
    doc_none = Doc(text="Text about science.")
    doc_none.results["clf"] = classification.ResultSingleLabel(label="politics", score=0.9)
    doc_none.gold["clf"] = "science"
    report_none = task.evaluate([doc_none])
    assert report_none.metrics[task.metric] == 0.0

    # 3. Multi-label partial overlap
    task_multi = Classification(labels=labels, model=batch_runtime.model, task_id="clf_multi", mode="multi")
    doc_partial = Doc(text="Text about science and politics.")
    # Prediction: science=0.9, politics=0.1
    # Gold: science=1.0, politics=1.0 (both are true)
    doc_partial.results["clf_multi"] = classification.ResultMultiLabel(label_scores=[("science", 0.9), ("politics", 0.1)])
    doc_partial.gold["clf_multi"] = classification.ResultMultiLabel(label_scores=[("science", 1.0), ("politics", 1.0)])

    report_partial = task_multi.evaluate([doc_partial])
    # Multi-label evaluation uses 1 - abs(gold - pred) averaged over labels
    # Label science: 1 - abs(1.0 - 0.9) = 0.9
    # Label politics: 1 - abs(1.0 - 0.1) = 0.1
    # Average: (0.9 + 0.1) / 2 = 0.5
    assert report_partial.metrics[task_multi.metric] == 0.5

    # 4. Multi-label full overlap
    doc_multi_full = Doc(text="Text about science and politics.")
    doc_multi_full.results["clf_multi"] = classification.ResultMultiLabel(
        label_scores=[("science", 1.0), ("politics", 1.0)]
    )
    doc_multi_full.gold["clf_multi"] = classification.ResultMultiLabel(
        label_scores=[("science", 1.0), ("politics", 1.0)]
    )
    report_multi_full = task_multi.evaluate([doc_multi_full])
    assert report_multi_full.metrics[task_multi.metric] == 1.0


@pytest.mark.parametrize("batch_runtime", [ModelType.huggingface], indirect=["batch_runtime"])
def test_list_str_multi_gold(batch_runtime) -> None:
    """Test using list of strings as gold labels for multi-label classification."""
    labels = ["science", "politics"]
    task_multi = Classification(labels=labels, model=batch_runtime.model, task_id="clf_multi", mode="multi")

    # 1. Full overlap with list[str] gold
    doc_full = Doc(text="Text about science and politics.")
    doc_full.results["clf_multi"] = classification.ResultMultiLabel(
        label_scores=[("science", 1.0), ("politics", 1.0)]
    )
    doc_full.gold["clf_multi"] = ["science", "politics"]
    report_full = task_multi.evaluate([doc_full])
    assert report_full.metrics[task_multi.metric] == 1.0

    # 2. Partial overlap with list[str] gold
    # Gold is only science. Pred is science and politics.
    doc_partial = Doc(text="Text about science.")
    doc_partial.results["clf_multi"] = classification.ResultMultiLabel(
        label_scores=[("science", 1.0), ("politics", 1.0)]
    )
    doc_partial.gold["clf_multi"] = ["science"]

    # Science: |1.0 - 1.0| = 0. Acc=1.0
    # Politics: |0.0 - 1.0| = 1. Acc=0.0
    # Avg: 0.5
    report_partial = task_multi.evaluate([doc_partial])
    assert report_partial.metrics[task_multi.metric] == 0.5



"""Aspect-based sentiment analysis."""

from sieves.tasks.predictive.schemas.sentiment_analysis import FewshotExample, Result
from sieves.tasks.predictive.sentiment_analysis.core import SentimentAnalysis

__all__ = ["SentimentAnalysis", "FewshotExample", "Result"]



"""
Test file containing examples for the Optimization guide.

These code blocks are referenced in docs/guides/optimization.md using snippet injection.

IMPORTANT: These tests use MINIMAL settings to reduce cost:
- num_trials=1 (instead of default 30)
- num_candidates=2 (instead of default 10)
- Only 4 training examples (instead of 20+)
"""

import pytest


def test_basic_optimization_example(small_dspy_model):
    """Test basic classification optimization with minimal settings."""
    model = small_dspy_model

    # --8<-- [start:optimization-classification-basic]
    # --8<-- [start:optimization-imports]
    import dspy
    from sieves import tasks, Doc
    from sieves.model_wrappers.utils import ModelSettings
    from sieves.tasks import Optimizer
    from sieves.tasks.predictive.classification import FewshotExampleSingleLabel
    # --8<-- [end:optimization-imports]

    # --8<-- [start:optimization-training-data]
    # 1. Create minimal training data (only 4 examples for speed)
    examples = [
        FewshotExampleSingleLabel(
            text="New smartphone released",
            label="technology",
            score=1.0
        ),
        FewshotExampleSingleLabel(
            text="Senate votes on bill",
            label="politics",
            score=1.0
        ),
        FewshotExampleSingleLabel(
            text="Football match results",
            label="sports",
            score=1.0
        ),
        FewshotExampleSingleLabel(
            text="Software update available",
            label="technology",
            score=1.0
        ),
    ]
    # --8<-- [end:optimization-training-data]

    # --8<-- [start:optimization-task-setup]
    # 2. Define task with few-shot examples
    task = tasks.Classification(
        labels={
            "technology": "Technology news, AI, software, and digital innovations",
            "politics": "Political events, elections, and government affairs",
            "sports": "Sports news, games, athletes, and competitions"
        },
        model=model,
        fewshot_examples=examples,
        mode='single',
        model_settings=ModelSettings(),
    )
    # --8<-- [end:optimization-task-setup]

    # --8<-- [start:optimization-optimizer-config]
    # 3. Create optimizer with MINIMAL settings for cost efficiency
    optimizer = Optimizer(
        model=model,
        val_frac=0.25,              # Use 25% for validation (1 example)
        seed=42,
        shuffle=True,
        dspy_init_kwargs=dict(
            auto=None,                  # Disable auto mode to use manual settings
            num_candidates=2,           # Minimal candidates (instead of 10)
            max_bootstrapped_demos=1,   # Minimal bootstrapped demos
            max_labeled_demos=1,        # Minimal labeled demos
            max_errors=10,              # Max errors before stopping
            num_threads=1               # Single thread
        ),
        dspy_compile_kwargs=dict(
            num_trials=1,               # Only 1 trial (instead of 30)
            minibatch=False
        )
    )
    # --8<-- [end:optimization-optimizer-config]

    # --8<-- [start:optimization-run]
    # 4. Run optimization
    best_prompt, best_examples = task.optimize(optimizer, verbose=True)

    print(f"Optimized prompt: {best_prompt}")
    print(f"Number of selected examples: {len(best_examples)}")
    # --8<-- [end:optimization-run]
    # --8<-- [end:optimization-classification-basic]

    assert best_prompt is not None



# mypy: ignore-errors
import pytest

from sieves import Doc, Pipeline
from sieves.model_wrappers import ModelType, ModelSettings, dspy_, langchain_, outlines_
from sieves.serialization import Config
from sieves.tasks import PredictiveTask, Summarization
from sieves.tasks.predictive import summarization
from sieves.tasks.predictive.schemas.summarization import Result


@pytest.mark.parametrize(
    "batch_runtime",
    Summarization.supports(),
    indirect=["batch_runtime"],
)
@pytest.mark.parametrize("fewshot", [True, False])
def test_run(summarization_docs, batch_runtime, fewshot) -> None:
    fewshot_examples = [
        summarization.FewshotExample(
            text="They counted: one, two, three, four, five, six, seven, eight, nine, ten, eleven, twelve, thirteen, "
            "fourteen.",
            n_words=6,
            summary="They counted from one to fourteen.",
            score=1.0,
        ),
        summarization.FewshotExample(
            text="Next in order were the Boeotians, led by Peneleos, Leitus, Arcesilaus, Prothoenor, and Clonius. "
            "These had with them fifty ships, and on board of each were a hundred and twenty young men of the "
            "Boeotians. Then came the men of Orchomenus, who lived in the realm of the Minyans, led by Ascalaphus"
            " and Ialmenus, sons of Mars. In their command were thirty ships. Next were the Phocians, led by"
            " Schedius and Epistrophus, sons of Iphitus the son of Naubolus. These had forty ships…",
            n_words=10,
            summary="Boeotians, Orchomenians, and Phocians sailed to Troy with many ships.",
            score=1.0,
        ),
    ]

    fewshot_args = {"fewshot_examples": fewshot_examples} if fewshot else {}
    task = summarization.Summarization(
        n_words=10,
        model=batch_runtime.model,
        model_settings=batch_runtime.model_settings,
        batch_size=batch_runtime.batch_size,
        **fewshot_args,
    )
    pipe = Pipeline(task)
    docs = list(pipe(summarization_docs))

    assert len(docs) == 2
    for doc in docs:
        assert doc.text
        assert "Summarization" in doc.results

        # Verify unified result types.
        assert isinstance(doc.results["Summarization"], summarization.Result)

        assert "Summarization" in doc.meta
        assert "raw" in doc.meta["Summarization"]
        assert "usage" in doc.meta["Summarization"]
        assert "usage" in doc.meta

        print(f"Output: {doc.results['Summarization']}")
        print(f"Raw output: {doc.meta['Summarization']['raw']}")
        print(f"Usage: {doc.meta['Summarization']['usage']}")
        print(f"Total Usage: {doc.meta['usage']}")

    with pytest.raises(NotImplementedError):
        pipe["Summarization"].distill(None, None, None, None, None, None, None, None)

    if fewshot:
        _to_hf_dataset(task, docs)


def _to_hf_dataset(task: Summarization, docs: list[Doc]) -> None:
    """Tests whether conversion to HF dataset works as expected.

    :param task: Summarization task instance.
    :param docs: List of documents to convert.
    """
    dataset = task.to_hf_dataset(docs)
    assert all([key in dataset.features for key in ("text", "summary")])
    assert len(dataset) == 2
    records = list(dataset)
    assert records[0]["text"].strip().startswith("The decay spreads over the State")
    assert records[1]["text"].strip().startswith("After all, the practical reason")
    for record in records:
        assert isinstance(record["summary"], str)

    with pytest.raises(KeyError):
        task.to_hf_dataset([Doc(text="This is a dummy text.")])


@pytest.mark.parametrize("batch_runtime", [ModelType.dspy], indirect=["batch_runtime"])
def test_serialization(summarization_docs, batch_runtime) -> None:
    pipe = Pipeline([
        summarization.Summarization(
            n_words=10,
            model=batch_runtime.model,
            model_settings=batch_runtime.model_settings,
            batch_size=batch_runtime.batch_size,
        )
    ])

    config = pipe.serialize()
    assert config.model_dump() == {'cls_name': 'sieves.pipeline.core.Pipeline',
 'tasks': {'is_placeholder': False,
           'value': [{'cls_name': 'sieves.tasks.predictive.summarization.core.Summarization',
                      'fewshot_examples': {'is_placeholder': False,
                                           'value': ()},
                      'batch_size': {'is_placeholder': False, "value": -1},
                      'model_settings': {'is_placeholder': False,
                                              'value': {
                                                        'config_kwargs': None,
                                                        'inference_kwargs': None,
                                                        'init_kwargs': None,
                                                        'strict': True, 'inference_mode': None}},
                      'include_meta': {'is_placeholder': False, 'value': True},
                      'model': {'is_placeholder': True,
                                'value': 'dspy.clients.lm.LM'},
                      'n_words': {'is_placeholder': False, 'value': 10},
                      'prompt_instructions': {'is_placeholder': False,
                                          'value': None},
                      'task_id': {'is_placeholder': False,
                                  'value': 'Summarization'},
                      'condition': {'is_placeholder': False, 'value': None},
                      'version': Config.get_version()}]},
 'use_cache': {'is_placeholder': False, 'value': True},
 'version': Config.get_version()}

    Pipeline.deserialize(config=config, tasks_kwargs=[{"model": batch_runtime.model}])


@pytest.mark.parametrize(
    "batch_runtime",
    Summarization.supports(),
    indirect=["batch_runtime"],
)
def test_inference_mode_override(batch_runtime) -> None:
    """Test that inference_mode parameter overrides the default value."""
    dummy = "dummy_inference_mode"

    task = summarization.Summarization(
        n_words=10,
        model=batch_runtime.model,
        model_settings=ModelSettings(inference_mode=dummy),
        batch_size=batch_runtime.batch_size,
    )

    assert task._bridge.inference_mode == dummy


@pytest.mark.parametrize("batch_runtime", [ModelType.dspy], indirect=["batch_runtime"])
def test_evaluation(batch_runtime) -> None:
    """Test evaluation for summarization using a real DSPy judge."""
    task = summarization.Summarization(n_words=10, model=batch_runtime.model, task_id="sum")

    # 1. Full overlap
    doc_full = Doc(text="The quick brown fox jumps over the lazy dog.")
    doc_full.results["sum"] = Result(summary="Fast fox jumps dog.", score=1.0)
    doc_full.gold["sum"] = Result(summary="Fast fox jumps dog.", score=1.0)
    report_full = task.evaluate([doc_full], judge=batch_runtime.model)
    assert report_full.metrics[task.metric] > 0.8

    # 2. No overlap
    doc_none = Doc(text="The quick brown fox jumps over the lazy dog.")
    doc_none.results["sum"] = Result(summary="The weather is nice today.", score=1.0)
    doc_none.gold["sum"] = Result(summary="Fast fox jumps dog.", score=1.0)
    report_none = task.evaluate([doc_none], judge=batch_runtime.model)
    assert report_none.metrics[task.metric] < 0.6

    # 3. Partial overlap
    doc_partial = Doc(text="The quick brown fox jumps over the lazy dog.")
    doc_partial.results["sum"] = Result(summary="A fox jumps.", score=1.0)
    doc_partial.gold["sum"] = Result(summary="The quick brown fox jumps over the lazy dog.", score=1.0)
    report_partial = task.evaluate([doc_partial], judge=batch_runtime.model)
    assert 0.2 <= report_partial.metrics[task.metric] <= 0.8



"""Base schemas and type variables for predictive tasks."""

from __future__ import annotations

import abc
from collections.abc import Sequence
from typing import Self, TypeVar

import dspy
import pydantic

from sieves.model_wrappers import ModelWrapper, ModelWrapperInferenceMode

TaskPromptSignature = TypeVar("TaskPromptSignature", covariant=True)
TaskResult = TypeVar("TaskResult")
TaskBridge = TypeVar("TaskBridge", bound="Bridge[TaskPromptSignature, TaskResult, ModelWrapperInferenceMode]")  # type: ignore[valid-type]  # noqa: F821


class EvaluationSignature(dspy.Signature):
    """Evaluate similarity between ground truth and predicted outputs."""

    target_fields: str = dspy.InputField(desc="Names of output fields being compared.")
    ground_truth: str = dspy.InputField(desc="Ground truth output values.")
    prediction: str = dspy.InputField(desc="Predicted output values.")

    similarity_score: float = dspy.OutputField(
        desc="Similarity score between 0.0 and 1.0, where 1.0 means identical and 0.0 means completely different."
    )


class FewshotExample(pydantic.BaseModel):
    """Few-shot example.

    Attributes:
        text: Input text.
    """

    text: str

    @property
    def input_fields(self) -> Sequence[str]:
        """Defines which fields are inputs.

        :return: Sequence of field names.
        """
        return ("text",)

    @property
    @abc.abstractmethod
    def target_fields(self) -> Sequence[str]:
        """Define which fields are targets, i.e. the end results the task aims to produce.

        :return: Sequence of field names.
        """

    def to_dspy(self) -> dspy.Example:
        """Convert to `dspy.Example`.

        :returns: Example as `dspy.Example`.
        """
        return dspy.Example(**ModelWrapper.convert_fewshot_examples([self])[0]).with_inputs(*self.input_fields)

    @classmethod
    def from_dspy(cls, example: dspy.Example) -> Self:
        """Convert from `dspy.Example`.

        :param example: Example as `dspy.Example`.
        :returns: Example as `FewshotExample`.
        """
        return cls(**example)



"""Schemas for classification task."""

from __future__ import annotations

import dspy
import gliner2.inference.engine
import pydantic

from sieves.model_wrappers import dspy_, gliner_, huggingface_, langchain_, outlines_
from sieves.tasks.predictive.schemas.core import FewshotExample as BaseFewshotExample


class FewshotExampleMultiLabel(BaseFewshotExample):
    """Few‑shot example for multi‑label classification with per‑label scores.

    Attributes:
        text: Input text.
        score_per_label: Mapping of labels to confidence scores.
    """

    score_per_label: dict[str, float]

    @pydantic.model_validator(mode="after")
    def check_score(self) -> FewshotExampleMultiLabel:
        """Validate that scores lie within [0, 1].

        :return: Validated instance.
        """
        if any([conf for conf in self.score_per_label.values() if not 0 <= conf <= 1]):
            raise ValueError("Score has to be between 0 and 1.")
        return self

    @property
    def target_fields(self) -> tuple[str, ...]:
        """Return target fields.

        :return: Target fields.
        """
        return ("score_per_label",)


class FewshotExampleSingleLabel(BaseFewshotExample):
    """Few‑shot example for single‑label classification with a global score.

    Attributes:
        text: Input text.
        label: Predicted label.
        score: Confidence score.
    """

    label: str
    score: float

    @pydantic.model_validator(mode="after")
    def check_score(self) -> FewshotExampleSingleLabel:
        """Check score value.

        :return: Validated instance.
        """
        if not (0 <= self.score <= 1):
            raise ValueError("Score has to be between 0 and 1.")
        return self

    @property
    def target_fields(self) -> tuple[str, ...]:
        """Return target fields.

        :return: Target fields.
        """
        return ("label", "score")


# --8<-- [start:Result]
class ResultSingleLabel(pydantic.BaseModel):
    """Result of a single-label classification task.

    Attributes:
        label: Predicted label.
        score: Confidence score.
    """

    label: str
    score: float


class ResultMultiLabel(pydantic.BaseModel):
    """Result of a multi-label classification task.

    Attributes:
        label_scores: List of label-score pairs.
    """

    label_scores: list[tuple[str, float]]


# --8<-- [end:Result]


TaskModel = dspy_.Model | gliner_.Model | langchain_.Model | huggingface_.Model | outlines_.Model
TaskPromptSignature = (
    type[dspy.Signature]
    | type[pydantic.BaseModel]
    | gliner2.inference.engine.Schema
    | gliner2.inference.engine.StructureBuilder
    | list[str]
)
TaskResult = ResultSingleLabel | ResultMultiLabel



"""Utils for model wrappers."""

import outlines
import transformers

from sieves.model_wrappers import (
    dspy_,
    gliner_,
    huggingface_,
    langchain_,
    outlines_,
)
from sieves.model_wrappers.core import (
    ModelWrapper,
    ModelWrapperInferenceMode,
    ModelWrapperModel,
    ModelWrapperPromptSignature,
    ModelWrapperResult,
)
from sieves.model_wrappers.types import ModelSettings

Model = dspy_.Model | gliner_.Model | huggingface_.Model | langchain_.Model | outlines_.Model


def init_default_model() -> outlines.models.Transformers:  # noqa: D401
    """Initialize default model (HuggingFaceTB/SmolLM-360M-Instruct with Outlines).

    :return: Initialized default model.
    """
    model_name = "HuggingFaceTB/SmolLM-360M-Instruct"

    return outlines.models.from_transformers(
        transformers.AutoModelForCausalLM.from_pretrained(model_name),
        transformers.AutoTokenizer.from_pretrained(model_name),
    )


def init_model_wrapper(
    model: Model, model_settings: ModelSettings
) -> ModelWrapper[ModelWrapperPromptSignature, ModelWrapperResult, ModelWrapperModel, ModelWrapperInferenceMode]:  # noqa: D401
    """Initialize internal model wrapper object.

    :param model: Model to use.
    :param model_settings: Settings for structured generation.
    :return ModelWrapper: ModelWrapper.
    :raises ValueError: If model type isn't supported.
    """
    model_type = type(model)
    module_wrapper_map = {
        dspy_: getattr(dspy_, "DSPy", None),
        gliner_: getattr(gliner_, "GliNER", None),
        huggingface_: getattr(huggingface_, "HuggingFace", None),
        langchain_: getattr(langchain_, "LangChain", None),
        outlines_: getattr(outlines_, "Outlines", None),
    }

    for module, model_wrapper_type in module_wrapper_map.items():
        assert hasattr(module, "Model")
        assert model_wrapper_type

        try:
            module_model_types = module.Model.__args__
        except AttributeError:
            module_model_types = (module.Model,)

        if any(issubclass(model_type, module_model_type) for module_model_type in module_model_types):
            internal_model_wrapper = model_wrapper_type(
                model=model,
                model_settings=model_settings,
            )
            assert isinstance(internal_model_wrapper, ModelWrapper)

            return internal_model_wrapper

    raise ValueError(
        f"Model type {model.__class__} is not supported. Please check the documentation and ensure that (1) you're "
        f"providing a supported model type and that (2) the corresponding library is installed in your environment."
    )



"""Hugging Face transformers zero-shot classification pipeline model wrapper."""

import enum
from collections.abc import Sequence
from typing import Any, override

import jinja2
import pydantic
import transformers

from sieves.model_wrappers.core import Executable, ModelWrapper
from sieves.model_wrappers.types import TokenUsage

PromptSignature = list[str]
Model = transformers.Pipeline
Result = dict[str, list[str] | list[float]]


class InferenceMode(enum.Enum):
    """Available inference modes."""

    zeroshot_cls = 0


class HuggingFace(ModelWrapper[PromptSignature, Result, Model, InferenceMode]):
    """ModelWrapper adapter around ``transformers.Pipeline`` for zero‑shot tasks."""

    @override
    @property
    def inference_modes(self) -> type[InferenceMode]:
        return InferenceMode

    @override
    @property
    def supports_few_shotting(self) -> bool:
        return True

    @override
    def build_executable(
        self,
        inference_mode: InferenceMode,
        prompt_template: str | None,
        prompt_signature: type[PromptSignature] | PromptSignature,
        fewshot_examples: Sequence[pydantic.BaseModel] = (),
    ) -> Executable[Result | None]:
        cls_name = self.__class__.__name__
        assert prompt_template, ValueError(f"prompt_template has to be provided to {cls_name} model wrapper by task.")
        assert isinstance(prompt_signature, list)

        # Render template with few-shot examples. Note that we don't use extracted document values here, as HF zero-shot
        # pipelines only support one hypothesis template per _call - and we want to batch, so our hypothesis template
        # will be document-invariant.
        fewshot_examples_dict = HuggingFace.convert_fewshot_examples(fewshot_examples)
        # Render hypothesis template with everything but text.
        template = jinja2.Template(prompt_template).render(**({"examples": fewshot_examples_dict}))

        def execute(values: Sequence[dict[str, Any]]) -> Sequence[tuple[Result | None, Any, TokenUsage]]:
            """Execute prompts with model wrapper for given values.

            :param values: Values to inject into prompts.
            :return: Sequence of tuples containing results, raw outputs, and token usage.
            """
            match inference_mode:
                case InferenceMode.zeroshot_cls:
                    results = self._model(
                        sequences=[doc_values["text"] for doc_values in values],
                        candidate_labels=prompt_signature,
                        hypothesis_template=template,
                        mode="multi",
                        **self._inference_kwargs,
                    )

                    # Estimate token usage if tokenizer is available.
                    tokenizer = self._get_tokenizer()

                    final_results: list[tuple[Result, Any, TokenUsage]] = []
                    for doc_values, res in zip(values, results):
                        usage = TokenUsage(
                            input_tokens=self._count_tokens(doc_values["text"], tokenizer),
                            # For classification, we estimate output tokens based on the labels.
                            output_tokens=self._count_tokens(" ".join(res["labels"]), tokenizer),
                        )

                        final_results.append((res, res, usage))
                    return final_results

                case _:
                    raise ValueError(f"Inference mode {inference_mode} not supported by {cls_name} model wrapper.")

        return execute

    @override
    def _get_tokenizer(self) -> Any | None:
        return getattr(self._model, "tokenizer", None)



"""Tasks."""

from sieves.tasks import predictive, preprocessing
from sieves.tasks.core import Task
from sieves.tasks.distillation.types import DistillationFramework
from sieves.tasks.optimization import Optimizer
from sieves.tasks.predictive import (
    NER,
    Classification,
    InformationExtraction,
    PIIMasking,
    QuestionAnswering,
    RelationExtraction,
    SentimentAnalysis,
    Summarization,
    Translation,
)
from sieves.tasks.predictive.core import PredictiveTask
from sieves.tasks.preprocessing import Chunking, Ingestion

__all__ = [
    "Chunking",
    "Classification",
    "DistillationFramework",
    "InformationExtraction",
    "Ingestion",
    "NER",
    "Optimizer",
    "SentimentAnalysis",
    "Summarization",
    "Translation",
    "QuestionAnswering",
    "PIIMasking",
    "Task",
    "RelationExtraction",
    "predictive",
    "PredictiveTask",
    "preprocessing",
]



"""Relation extraction predictive task."""

from __future__ import annotations

import warnings
from collections.abc import Callable, Iterable, Sequence
from pathlib import Path
from typing import Any, Literal, override

import datasets
import dspy
import pydantic

from sieves.data import Doc
from sieves.model_wrappers import ModelType, gliner_
from sieves.model_wrappers.types import ModelSettings
from sieves.serialization import Config
from sieves.tasks.distillation.types import DistillationFramework
from sieves.tasks.predictive.core import PredictiveTask
from sieves.tasks.predictive.gliner_bridge import GliNERBridge
from sieves.tasks.predictive.relation_extraction.bridges import DSPyRelationExtraction, PydanticRelationExtraction
from sieves.tasks.predictive.schemas.relation_extraction import (
    FewshotExample,
    TaskModel,
    TaskPromptSignature,
    TaskResult,
)

_TaskBridge = GliNERBridge | DSPyRelationExtraction | PydanticRelationExtraction


class RelationExtraction(PredictiveTask[TaskPromptSignature, TaskResult, _TaskBridge]):
    """Extract relations between entities in text."""

    def __init__(
        self,
        relations: Sequence[str] | dict[str, str],
        model: TaskModel,
        entity_types: Sequence[str] | dict[str, str] | None = None,
        task_id: str | None = None,
        include_meta: bool = True,
        batch_size: int = -1,
        prompt_instructions: str | None = None,
        fewshot_examples: Sequence[FewshotExample] = (),
        model_settings: ModelSettings = ModelSettings(),
        condition: Callable[[Doc], bool] | None = None,
    ) -> None:
        """Initialize RelationExtraction task.

        :param relations: Relations to extract. Can be a list of relation types or a dict mapping types to descriptions.
        :param model: Model to use.
        :param entity_types: Optional constraints on entity types involved in relations.
        :param task_id: Task ID.
        :param include_meta: Whether to include meta information generated by the task.
        :param batch_size: Batch size to use for inference. Use -1 to process all documents at once.
        :param prompt_instructions: Custom prompt instructions. If None, default instructions are used.
        :param fewshot_examples: Few-shot examples.
        :param model_settings: Settings for structured generation.
        :param condition: Optional callable that determines whether to process each document.
        """
        self._relations = relations
        self._entity_types = entity_types

        super().__init__(
            model=model,
            task_id=task_id,
            include_meta=include_meta,
            batch_size=batch_size,
            overwrite=False,
            prompt_instructions=prompt_instructions,
            fewshot_examples=fewshot_examples,
            model_settings=model_settings,
            condition=condition,
        )

    @property
    @override
    def fewshot_example_type(self) -> type[FewshotExample]:
        """Return few-shot example type.

        :return: Few-shot example type.
        """
        return FewshotExample

    @property
    @override
    def prompt_signature(self) -> type[pydantic.BaseModel]:
        """Return the unified Pydantic prompt signature for this task.

        :return: Unified Pydantic prompt signature.
        """
        # Define relation type literal if relations are provided.
        if isinstance(self._relations, dict):
            relation_names = list(self._relations.keys())
        else:
            relation_names = list(self._relations)

        RelationType = Literal[*(tuple(relation_names))] if relation_names else str  # type: ignore[invalid-type-form]

        # Define entity type literal if entity types are provided.
        entity_type_names: list[str] = []
        if self._entity_types:
            if isinstance(self._entity_types, dict):
                entity_type_names = list(self._entity_types.keys())
            else:
                entity_type_names = list(self._entity_types)

        EntityType = Literal[*(tuple(entity_type_names))] if entity_type_names else str  # type: ignore[invalid-type-form]

        # Create dynamic models.
        DynamicEntity = pydantic.create_model(
            "RelationEntity",
            text=(str, pydantic.Field(..., description="Surface text of the entity as it appears in the document.")),
            entity_type=(
                EntityType,
                pydantic.Field(..., description="The category or type of the entity."),
            ),
            __doc__="An entity involved in a relation.",
            __base__=pydantic.BaseModel,
        )

        DynamicTriplet = pydantic.create_model(
            "RelationTriplet",
            head=(DynamicEntity, pydantic.Field(..., description="The subject entity (head) of the relation.")),
            relation=(RelationType, pydantic.Field(..., description="The type of relation between the head and tail.")),
            tail=(DynamicEntity, pydantic.Field(..., description="The object entity (tail) of the relation.")),
            score=(
                float | None,
                pydantic.Field(
                    default=None, description="Provide a confidence score for this relation triplet, between 0 and 1."
                ),
            ),
            __doc__="A relation triplet consisting of a head entity, a relation type, and a tail entity.",
            __base__=pydantic.BaseModel,
        )

        return pydantic.create_model(
            "RelationExtractionOutput",
            __doc__="Result of relation extraction. Contains a list of extracted relation triplets.",
            triplets=(
                list[DynamicTriplet],  # type: ignore[invalid-type-form]
                pydantic.Field(..., description="List of extracted relation triplets."),
            ),
        )

    @property
    @override
    def metric(self) -> str:
        return "F1"

    @override
    def _compute_metrics(self, truths: list[Any], preds: list[Any], judge: dspy.LM | None = None) -> dict[str, float]:
        """Compute corpus-level metrics.

        :param truths: List of ground truths.
        :param preds: List of predictions.
        :param judge: Optional DSPy LM instance to use as judge for generative tasks.
        :return: Dictionary of metrics.
        """
        tp = 0
        fp = 0
        fn = 0

        for gold, pred in zip(truths, preds):
            if gold is not None:
                assert isinstance(gold, TaskResult)
                true_triplets = {(t.head.text.lower(), t.relation.lower(), t.tail.text.lower()) for t in gold.triplets}
            else:
                true_triplets = set()

            if pred is not None:
                assert isinstance(pred, TaskResult)
                pred_triplets = {(t.head.text.lower(), t.relation.lower(), t.tail.text.lower()) for t in pred.triplets}
            else:
                pred_triplets = set()

            tp += len(true_triplets & pred_triplets)
            fp += len(pred_triplets - true_triplets)
            fn += len(true_triplets - pred_triplets)

        precision = tp / (tp + fp) if (tp + fp) > 0 else 0.0
        recall = tp / (tp + fn) if (tp + fn) > 0 else 0.0
        f1 = 2 * precision * recall / (precision + recall) if (precision + recall) > 0 else 0.0

        return {self.metric: f1}

    @override
    def _init_bridge(self, model_type: ModelType) -> _TaskBridge:
        if model_type == ModelType.gliner:
            if self._entity_types is not None:
                warnings.warn(
                    "GliNER2 backend does not support entity type constraints for relation extraction. "
                    "The `entity_types` parameter will be ignored.",
                )

            return GliNERBridge(
                task_id=self._task_id,
                prompt_instructions=self._custom_prompt_instructions,
                prompt_signature=self.prompt_signature,
                model_settings=self._model_settings,
                inference_mode=gliner_.InferenceMode.relations,
            )

        bridge_types: dict[ModelType, type[DSPyRelationExtraction | PydanticRelationExtraction]] = {
            ModelType.dspy: DSPyRelationExtraction,
            ModelType.langchain: PydanticRelationExtraction,
            ModelType.outlines: PydanticRelationExtraction,
        }

        try:
            return bridge_types[model_type](
                task_id=self._task_id,
                relations=self._relations,
                entity_types=self._entity_types,
                prompt_instructions=self._custom_prompt_instructions,
                model_settings=self._model_settings,
                prompt_signature=self.prompt_signature,
                model_type=model_type,
                fewshot_examples=self._fewshot_examples,
            )
        except KeyError as err:
            raise KeyError(f"Model type {model_type} is not supported by {self.__class__.__name__}.") from err

    @staticmethod
    @override
    def supports() -> set[ModelType]:
        return {
            ModelType.dspy,
            ModelType.gliner,
            ModelType.langchain,
            ModelType.outlines,
        }

    @override
    @property
    def _state(self) -> dict[str, Any]:
        return {
            **super()._state,
            "relations": self._relations,
            "entity_types": self._entity_types,
        }

    @override
    def to_hf_dataset(self, docs: Iterable[Doc], threshold: float | None = None) -> datasets.Dataset:
        # Define metadata and features.
        entity_feature = datasets.Features(
            {
                "text": datasets.Value("string"),
                "entity_type": datasets.Value("string"),
            }
        )
        triplet_feature = datasets.Features(
            {
                "head": entity_feature,
                "relation": datasets.Value("string"),
                "tail": entity_feature,
                "score": datasets.Value("float32"),
            }
        )
        features = datasets.Features({"text": datasets.Value("string"), "triplets": datasets.Sequence(triplet_feature)})
        info = datasets.DatasetInfo(
            description=f"Relation extraction dataset. Generated with sieves v{Config.get_version()}.",
            features=features,
        )

        # Fetch data used for generating dataset.
        try:
            data: list[dict[str, Any]] = []
            for doc in docs:
                triplets: list[dict[str, Any]] = []
                for triplet in doc.results[self._task_id].triplets:
                    triplets.append(
                        {
                            "head": triplet.head.model_dump(),
                            "relation": triplet.relation,
                            "tail": triplet.tail.model_dump(),
                            "score": triplet.score,
                        }
                    )
                data.append({"text": doc.text, "triplets": triplets})
        except KeyError as err:
            raise KeyError(f"Not all documents have results for this task with ID {self._task_id}") from err

        # Create dataset.
        return datasets.Dataset.from_list(data, features=features, info=info)

    @override
    def distill(
        self,
        base_model_id: str,
        framework: DistillationFramework,
        data: datasets.Dataset | Sequence[Doc],
        output_path: Path | str,
        val_frac: float,
        init_kwargs: dict[str, Any] | None = None,
        train_kwargs: dict[str, Any] | None = None,
        seed: int | None = None,
    ) -> None:
        raise NotImplementedError

    @override
    def _evaluate_dspy_example(self, truth: dspy.Example, pred: dspy.Prediction, trace: Any, model: dspy.LM) -> float:
        # Compute triplet-level F1 score based on (head_text, relation, tail_text) triples.
        # Use lowercase for robust matching.
        true_triplets = {
            (t["head"]["text"].lower(), t["relation"].lower(), t["tail"]["text"].lower()) for t in truth["triplets"]
        }
        pred_triplets = {
            (t["head"]["text"].lower(), t["relation"].lower(), t["tail"]["text"].lower())
            for t in pred.get("triplets", [])
        }

        if not true_triplets:
            base_f1 = 1.0 if not pred_triplets else 0.0
        else:
            precision = len(true_triplets & pred_triplets) / len(pred_triplets) if pred_triplets else 0
            recall = len(true_triplets & pred_triplets) / len(true_triplets)
            base_f1 = 2 * precision * recall / (precision + recall) if (precision + recall) > 0 else 0

        # If score is available, incorporate it.
        if "triplets" in truth and "triplets" in pred:
            true_scores = [t.get("score") for t in truth["triplets"] if t.get("score") is not None]
            pred_scores = [t.get("score") for t in pred["triplets"] if t.get("score") is not None]

            if true_scores and pred_scores:
                true_avg = sum(true_scores) / len(true_scores)
                pred_avg = sum(pred_scores) / len(pred_scores)
                score_accuracy = 1 - abs(true_avg - max(min(pred_avg, 1), 0))
                return (base_f1 + score_accuracy) / 2

        return base_f1



"""Schemas for information extraction task."""

from __future__ import annotations

import dspy
import gliner2.inference.engine
import pydantic

from sieves.model_wrappers import dspy_, gliner_, langchain_, outlines_
from sieves.tasks.predictive.schemas.core import FewshotExample as BaseFewshotExample


class FewshotExampleMulti(BaseFewshotExample):
    """Few-shot example for multi-entity extraction.

    Attributes:
        text: Input text.
        entities: List of entities.
    """

    entities: list[pydantic.BaseModel]

    @property
    def target_fields(self) -> tuple[str, ...]:
        """Return target fields.

        :return: Target fields.
        """
        return ("entities",)


class FewshotExampleSingle(BaseFewshotExample):
    """Few-shot example for single-entity extraction.

    Attributes:
        text: Input text.
        entity: Extracted entity.
    """

    entity: pydantic.BaseModel | None

    @property
    def target_fields(self) -> tuple[str, ...]:
        """Return target fields.

        :return: Target fields.
        """
        return ("entity",)


# --8<-- [start:Result]
class ResultSingle(pydantic.BaseModel):
    """Result of a single-entity extraction task.

    Attributes:
        entity: Extracted entity.
    """

    entity: pydantic.BaseModel | None


class ResultMulti(pydantic.BaseModel):
    """Result of a multi-entity extraction task.

    Attributes:
        entities: List of extracted entities.
    """

    entities: list[pydantic.BaseModel]


# --8<-- [end:Result]


TaskModel = dspy_.Model | gliner_.Model | langchain_.Model | outlines_.Model
TaskPromptSignature = (
    type[dspy.Signature]
    | type[pydantic.BaseModel]
    | gliner2.inference.engine.Schema
    | gliner2.inference.engine.StructureBuilder
)
TaskResult = ResultSingle | ResultMulti



"""NER task."""

from sieves.tasks.predictive.ner.core import NER
from sieves.tasks.predictive.schemas.ner import (
    FewshotExample,
    Result,
    TaskPromptSignature,
    TaskResult,
)

__all__ = ["NER", "FewshotExample", "Result", "TaskResult", "TaskPromptSignature"]



"""Bridges for relation extraction task."""

from __future__ import annotations

import abc
from collections.abc import Callable, Iterable, Sequence
from typing import Any, TypeVar, override

import pydantic

from sieves.data import Doc
from sieves.model_wrappers import (
    ModelType,
    ModelWrapperInferenceMode,
    dspy_,
    langchain_,
    outlines_,
)
from sieves.model_wrappers.types import ModelSettings
from sieves.tasks.predictive.bridges import Bridge
from sieves.tasks.predictive.consolidation import MultiEntityConsolidation
from sieves.tasks.predictive.schemas.relation_extraction import (
    RelationEntity,
    RelationTriplet,
    Result,
)
from sieves.tasks.predictive.utils import convert_to_signature

_BridgePromptSignature = TypeVar("_BridgePromptSignature")
_BridgeResult = TypeVar("_BridgeResult")


class RelationExtractionBridge(Bridge[_BridgePromptSignature, _BridgeResult, ModelWrapperInferenceMode], abc.ABC):
    """Abstract base class for relation extraction bridges."""

    def __init__(
        self,
        task_id: str,
        relations: Sequence[str] | dict[str, str],
        entity_types: Sequence[str] | dict[str, str] | None,
        prompt_instructions: str | None,
        model_settings: ModelSettings,
        prompt_signature: type[pydantic.BaseModel],
        model_type: ModelType,
        fewshot_examples: Sequence[pydantic.BaseModel] = (),
    ):
        """Initialize relation extraction bridge.

        :param task_id: Task ID.
        :param relations: Relations to extract. Can be a list of relation types or a dict mapping types to descriptions.
        :param entity_types: Entity types constraints.
        :param prompt_instructions: Custom prompt instructions. If None, default instructions are used.
        :param model_settings: Settings for structured generation.
        :param prompt_signature: Unified Pydantic prompt signature.
        :param model_type: Model type.
        :param fewshot_examples: Few-shot examples.
        """
        super().__init__(
            task_id=task_id,
            prompt_instructions=prompt_instructions,
            overwrite=False,
            model_settings=model_settings,
            prompt_signature=prompt_signature,
            model_type=model_type,
            fewshot_examples=fewshot_examples,
        )
        if isinstance(relations, dict):
            self._relations = list(relations.keys())
            self._relation_descriptions = relations
        else:
            self._relations = list(relations)
            self._relation_descriptions = {}

        if isinstance(entity_types, dict):
            self._entity_types = list(entity_types.keys())
            self._entity_type_descriptions = entity_types
        elif entity_types is not None:
            self._entity_types = list(entity_types)
            self._entity_type_descriptions = {}
        else:
            self._entity_types = None
            self._entity_type_descriptions = {}

        self._consolidation_strategy = MultiEntityConsolidation(extractor=self._chunk_extractor)

    @override
    @property
    def prompt_signature(self) -> _BridgePromptSignature:
        return convert_to_signature(
            model_cls=self._pydantic_signature,
            model_type=self.model_type,
            mode="relations",
        )  # type: ignore[return-value]

    @property
    @abc.abstractmethod
    def _chunk_extractor(self) -> Callable[[Any], Iterable[pydantic.BaseModel]]:
        """Return a callable that extracts a list of entities from a raw chunk result.

        :return: Extractor callable.
        """

    def _get_relation_descriptions(self) -> str:
        """Return relation descriptions as a string.

        :return: Relation descriptions.
        """
        descs: list[str] = []
        for rel in self._relations:
            if rel in self._relation_descriptions:
                descs.append(
                    f"  <relation_description>\n    <relation>{rel}</relation>\n    <description>"
                    f"{self._relation_descriptions[rel]}</description>\n  </relation_description>"
                )
            else:
                descs.append(f"  <relation>{rel}</relation>")
        return "\n".join(descs)

    def _get_entity_type_descriptions(self) -> str:
        """Return entity type descriptions as a string.

        :return: Entity type descriptions.
        """
        if self._entity_types is None:
            return "Unbounded"

        descs: list[str] = []
        for et in self._entity_types:
            if et in self._entity_type_descriptions:
                descs.append(
                    f"  <entity_type_description>\n    <type>{et}</type>\n    <description>"
                    f"{self._entity_type_descriptions[et]}</description>\n  </entity_type_description>"
                )
            else:
                descs.append(f"  <type>{et}</type>")

        return "\n".join(descs)

    def _process_triplets(self, raw_triplets: list[Any]) -> list[RelationTriplet]:
        """Convert raw triplets from model to RelationTriplet objects.

        :param raw_triplets: Raw triplets from the model.
        :return: Processed RelationTriplet objects.
        """
        processed: list[RelationTriplet] = []
        for raw in raw_triplets:
            head_text = getattr(raw.head, "text", "")
            head_type = getattr(raw.head, "entity_type", "")

            tail_text = getattr(raw.tail, "text", "")
            tail_type = getattr(raw.tail, "entity_type", "")

            processed.append(
                RelationTriplet(
                    head=RelationEntity(text=head_text, entity_type=head_type),
                    relation=getattr(raw, "relation", ""),
                    tail=RelationEntity(text=tail_text, entity_type=tail_type),
                    score=getattr(raw, "score", None),
                )
            )

        return processed

    @override
    def integrate(self, results: Sequence[_BridgeResult | list[Any]], docs: list[Doc]) -> list[Doc]:
        for doc, result in zip(docs, results):
            # Handle both model result objects and raw lists from consolidation.
            raw_triplets = result if isinstance(result, list) else getattr(result, "triplets", [])
            doc.results[self._task_id] = Result(triplets=self._process_triplets(raw_triplets))

        return docs

    @override
    def consolidate(
        self,
        results: Sequence[_BridgeResult],
        docs_offsets: list[tuple[int, int]],
    ) -> Sequence[list[pydantic.BaseModel]]:
        return self._consolidation_strategy.consolidate(results, docs_offsets)


class DSPyRelationExtraction(RelationExtractionBridge[dspy_.PromptSignature, dspy_.Result, dspy_.InferenceMode]):
    """DSPy bridge for relation extraction."""

    @override
    def _validate(self) -> None:
        assert self._model_type == ModelType.dspy

    @property
    @override
    def _chunk_extractor(self) -> Callable[[Any], Iterable[pydantic.BaseModel]]:
        return lambda res: res.triplets

    @override
    @property
    def _default_prompt_instructions(self) -> str:
        return ""

    @override
    @property
    def inference_mode(self) -> dspy_.InferenceMode:
        return self._model_settings.inference_mode or dspy_.InferenceMode.predict


class PydanticRelationExtraction(
    RelationExtractionBridge[pydantic.BaseModel, pydantic.BaseModel | list[Any], ModelWrapperInferenceMode], abc.ABC
):
    """Base class for Pydantic-based relation extraction bridges."""

    @override
    def _validate(self) -> None:
        assert self._model_type in {ModelType.langchain, ModelType.outlines}

    @property
    @override
    def _chunk_extractor(self) -> Callable[[Any], Iterable[pydantic.BaseModel]]:
        return lambda res: res.triplets

    @override
    @property
    def _default_prompt_instructions(self) -> str:
        return (
            "Extract relations between entities in the text.\n"
            f"Relations: {self._relations}\n"
            f"Entity Types: {self._entity_types or 'Any'}\n"
            "Return a list of triplets with head, relation, tail, and a confidence score between 0.0 and 1.0."
        )

    @override
    @property
    def _prompt_conclusion(self) -> str | None:
        return "========\n<text>{{ text }}</text>"

    @override
    @property
    def model_type(self) -> ModelType:
        return self._model_type

    @override
    @property
    def inference_mode(self) -> outlines_.InferenceMode | langchain_.InferenceMode:
        if self._model_type == ModelType.outlines:
            return self._model_settings.inference_mode or outlines_.InferenceMode.json
        elif self._model_type == ModelType.langchain:
            return self._model_settings.inference_mode or langchain_.InferenceMode.structured

        raise ValueError(f"Unsupported model type: {self._model_type}")



# mypy: ignore-errors
import pytest

from sieves import Doc
from sieves.model_wrappers import ModelType
from sieves.tasks.predictive import SentimentAnalysis, sentiment_analysis
from sieves.tasks.predictive.schemas.sentiment_analysis import Result


@pytest.mark.parametrize(
    "batch_runtime",
    SentimentAnalysis.supports(),
    indirect=["batch_runtime"],
)
@pytest.mark.parametrize("fewshot", [True, False])
def test_run(sentiment_analysis_docs, batch_runtime, fewshot):
    fewshot_examples = [
        sentiment_analysis.FewshotExample(
            text="Beautiful dishes, haven't eaten so well in a long time.",
            sentiment_per_aspect={"overall": 1.0, "food": 1.0, "service": 0.5},
            score=1.0,
        ),
        sentiment_analysis.FewshotExample(
            text="Horrible place. Service is unfriendly, food overpriced and bland.",
            sentiment_per_aspect={"overall": 0.0, "food": 0.0, "service": 0.0},
            score=1.0,
        ),
    ]

    task = SentimentAnalysis(
        aspects=["food", "service", "overall"],
        model=batch_runtime.model,
        model_settings=batch_runtime.model_settings,
        fewshot_examples=fewshot_examples if fewshot else [],
        batch_size=batch_runtime.batch_size,
    )

    results = list(task(sentiment_analysis_docs))

    assert len(results) == 2
    for doc in results:
        assert "SentimentAnalysis" in doc.results
        res = doc.results["SentimentAnalysis"]

        assert isinstance(res, sentiment_analysis.Result)
        assert len(res.sentiment_per_aspect) == 3
        for aspect, score in res.sentiment_per_aspect.items():
            assert aspect in ["food", "service", "overall"]
            assert isinstance(score, float)

        print(f"Output: {doc.results['SentimentAnalysis']}")
        print(f"Raw output: {doc.meta['SentimentAnalysis']['raw']}")
        print(f"Usage: {doc.meta['SentimentAnalysis']['usage']}")
        print(f"Total Usage: {doc.meta['usage']}")


@pytest.mark.parametrize("batch_runtime", [ModelType.outlines], indirect=["batch_runtime"])
def test_evaluation(batch_runtime) -> None:
    """Test evaluation for sentiment analysis without running pipeline."""
    task = SentimentAnalysis(aspects=["food", "service"], model=batch_runtime.model, task_id="sent")

    # 1. Full overlap
    doc_full = Doc(text="Great food and service.")
    res_full = Result(sentiment_per_aspect={"food": 1.0, "service": 1.0, "overall": 1.0})
    doc_full.results["sent"] = res_full
    doc_full.gold["sent"] = res_full
    report_full = task.evaluate([doc_full])
    assert report_full.metrics[task.metric] == 1.0

    # 2. No overlap (opposite sentiments)
    doc_none = Doc(text="Great food and service.")
    res_none_pred = Result(sentiment_per_aspect={"food": 0.0, "service": 0.0, "overall": 0.0})
    doc_none.results["sent"] = res_none_pred
    doc_none.gold["sent"] = res_full
    report_none = task.evaluate([doc_none])
    assert report_none.metrics[task.metric] == 0.0

    # 3. Partial overlap
    doc_partial = Doc(text="Great food. Bad service.")
    # food is correct (1.0), service is wrong (0.0 vs 1.0)
    # F1 (Macro):
    # food: 1.0 (TP) -> F1 1.0
    # service: 0.0 (FN) -> F1 0.0
    # overall: 1.0 (TP) -> F1 1.0
    # Average: (1.0 + 0.0 + 1.0) / 3 = 0.66...
    res_partial_pred = Result(sentiment_per_aspect={"food": 1.0, "service": 0.0, "overall": 1.0})
    doc_partial.results["sent"] = res_partial_pred
    doc_partial.gold["sent"] = res_full
    report_partial = task.evaluate([doc_partial])
    assert 0.6 < report_partial.metrics[task.metric] < 0.7



# mypy: ignore-errors
from pathlib import Path

import pytest
from marker.converters.pdf import PdfConverter
from marker.models import create_model_dict

from sieves import Doc, Pipeline, tasks
from sieves.serialization import Config


@pytest.mark.skip(reason="Currently running into OOM issues with instantiating Marker converts.")
def test_marker():
    """Workaround to keep memory usage low: run single function with one instantiated Marker instance."""
    marker_converter = PdfConverter(artifact_dict=create_model_dict())

    def test_run() -> None:
        resources = [Doc(uri=Path(__file__).parent.parent.parent / "assets" / "1204.0162v2.pdf")]
        pipe = Pipeline(tasks=[tasks.preprocessing.Marker(converter=marker_converter)])
        docs = list(pipe(resources))

        assert len(docs) == 1
        assert docs[0].text

    def test_with_extract_images() -> None:
        resources = [Doc(uri=Path(__file__).parent.parent.parent / "assets" / "1204.0162v2.pdf")]
        pipe = Pipeline(
            tasks=[tasks.preprocessing.Marker(converter=marker_converter, extract_images=True, include_meta=True)]
        )
        docs = list(pipe(resources))

        assert len(docs) == 1
        assert docs[0].text
        assert docs[0].images

    def test_serialization() -> None:
        resources = [Doc(uri=Path(__file__).parent.parent.parent / "assets" / "1204.0162v2.pdf")]
        pipe = Pipeline(tasks=[tasks.preprocessing.Marker(converter=marker_converter, include_meta=True)])
        docs = list(pipe(resources))

        config = pipe.serialize()
        version = Config.get_version()
        assert config.model_dump() == {
            "cls_name": "sieves.pipeline.core.Pipeline",
            "tasks": {
                "is_placeholder": False,
                "value": [
                    {
                        "cls_name": "sieves.tasks.preprocessing.marker_.Marker",
                        "converter": {"is_placeholder": True, "value": "marker.converters.pdf.PdfConverter"},
                        "export_format": {"is_placeholder": False, "value": "markdown"},
                        "extract_images": {"is_placeholder": False, "value": False},
                        "include_meta": {"is_placeholder": False, "value": True},
                        "task_id": {"is_placeholder": False, "value": "Marker"},
                        "version": version,
                    }
                ],
            },
            "version": version,
        }

        # For deserialization, we need to provide the converter
        converter = marker_converter
        deserialized_pipeline = Pipeline.deserialize(
            config=config, tasks_kwargs=[{"converter": converter, "export_format": "markdown"}]
        )
        deserialized_docs = list(deserialized_pipeline(resources))

        assert len(deserialized_docs) == 1
        assert deserialized_docs[0].text == docs[0].text

    test_run()
    test_with_extract_images()
    test_serialization()



"""Classification task."""

from sieves.tasks.predictive.classification.core import Classification
from sieves.tasks.predictive.schemas.classification import (
    FewshotExampleMultiLabel,
    FewshotExampleSingleLabel,
    ResultMultiLabel,
    ResultSingleLabel,
)

__all__ = [
    "Classification",
    "FewshotExampleMultiLabel",
    "FewshotExampleSingleLabel",
    "ResultMultiLabel",
    "ResultSingleLabel",
]



"""Wrapper for `Docling` for the conversion of complex files into markdown."""

import itertools
import sys
import warnings
from collections.abc import Callable, Iterable
from typing import Any, Literal

import docling.datamodel.document
import docling.datamodel.settings
import docling.document_converter
from loguru import logger

from sieves.data.doc import Doc
from sieves.tasks.core import Task

Converter = docling.document_converter.DocumentConverter


class Docling(Task):
    """Parser wrapping the docling library to convert files into documents."""

    def __init__(
        self,
        converter: Converter | None = None,
        export_format: Literal["markdown", "html"] = "markdown",
        task_id: str | None = None,
        include_meta: bool = False,
        batch_size: int = -1,
        condition: Callable[[Doc], bool] | None = None,
    ):
        """Initialize the docling parser.

        :param converter: Docling parser instance.
        :param task_id: Task ID.
        :param include_meta: Whether to include meta information generated by the task.
        :param batch_size: Batch size to use for processing. Use -1 to process all documents at once.
        :param condition: Optional callable that determines whether to process each document.
        """
        super().__init__(task_id=task_id, include_meta=include_meta, batch_size=batch_size, condition=condition)
        self._converter = converter if converter else docling.document_converter.DocumentConverter()
        self._export_format = export_format

        # Set batch size in Docling performance settings.
        docling.datamodel.settings.settings.perf.doc_batch_size = (
            self._batch_size if self._batch_size > 0 else sys.maxsize
        )

    def _validate_docs(self, docs: Iterable[Doc]) -> None:
        """Validate documents.

        Raises warning if docs already have text values, as they will be overwritten.

        :param docs: Docs to validate.
        :raises ValueError: If documents do not have a URI.
        """
        have_text = False
        for doc in docs:
            assert doc.uri, ValueError("Documents have to have a value for .uri.")
            if doc.text:
                have_text = True
        if have_text:
            warnings.warn(f"Task {self._task_id} is about to overwrite existing .text values.")

    def _call(self, docs: Iterable[Doc]) -> Iterable[Doc]:
        """Parse resources using docling.

        :param docs: Resources to process.
        :return: Parsed documents.
        :raises ValueError: If documents failed to parse.
        """
        docs_iters = itertools.tee(docs, 3)
        self._validate_docs(docs_iters[0])

        for doc, parsed_resource in zip(
            docs_iters[2], self._converter.convert_all([resource.uri for resource in docs_iters[1]])
        ):
            try:
                if self._include_meta:
                    doc.meta |= {self.id: parsed_resource}
                if self._export_format == "html":
                    doc.text = parsed_resource.document.export_to_html()
                else:
                    doc.text = parsed_resource.document.export_to_markdown()

            except Exception as e:
                logger.error(f"Failed to parse file {doc.uri}: {str(e)}")
                raise e

            yield doc

    @property
    def _state(self) -> dict[str, Any]:
        return {
            **super()._state,
            "converter": self._converter,
            "export_format": self._export_format,
        }



"""Question answering task."""

from sieves.tasks.predictive.question_answering.core import QuestionAnswering
from sieves.tasks.predictive.schemas.question_answering import FewshotExample, QuestionAnswer, Result

__all__ = ["QuestionAnswering", "FewshotExample", "QuestionAnswer", "Result"]



# mypy: ignore-errors
import datasets
import pydantic
import transformers

from sieves import tasks
from sieves.tasks.utils import PydanticToHFDatasets


def test_custom_prompt_instructions() -> None:
    prompt_instructions = "This is a different prompt template."
    task = tasks.predictive.Classification(
        task_id="classifier",
        labels=["science", "politics"],
        model=transformers.pipeline(
            "zero-shot-classification", model="MoritzLaurer/xtremedistil-l6-h256-zeroshot-v1.1-all-33"
        ),
        prompt_instructions=prompt_instructions,
    )
    assert task.prompt_template.strip().startswith(prompt_instructions)


def test_custom_prompt_signature_desc() -> None:
    prompt_instructions = "This is a different prompt signature description."
    task = tasks.predictive.Classification(
        task_id="classifier",
        labels=["science", "politics"],
        model=transformers.pipeline(
            "zero-shot-classification", model="MoritzLaurer/xtremedistil-l6-h256-zeroshot-v1.1-all-33"
        ),
        prompt_instructions=prompt_instructions,
    )
    assert task.prompt_template.strip().startswith(prompt_instructions)


def test_pydantic_to_hf() -> None:
    """Test the conversion of various Pydantic objects to HF datasets.Features."""

    # Check non-nested properties.

    class Simple(pydantic.BaseModel):
        a: int
        b: str
        c: str | float
        d: tuple[int, float]
        e: str | None
        f: str | None  # noqa: UP007

    features = PydanticToHFDatasets.model_cls_to_features(Simple)
    assert all([key in features for key in ("a", "b")])
    assert features["a"].dtype == "int32"
    assert features["b"].dtype == "string"
    assert features["c"].dtype == "string"
    assert isinstance(features["d"], datasets.Sequence)
    assert features["d"].feature.dtype == "string"
    assert features["e"].dtype == "string"
    assert features["f"].dtype == "string"
    assert PydanticToHFDatasets.model_to_dict(None) is None
    dataset = datasets.Dataset.from_list(
        [PydanticToHFDatasets.model_to_dict(Simple(a=1, b="blub", c=0.3, d=(1, 0.4), e=None, f=None))],
        features=features,
    )
    assert list(dataset)[0] == {"a": 1, "b": "blub", "c": "0.3", "d": ["1", "0.4"], "e": None, "f": None}

    # With a list of primitives.

    class WithList(pydantic.BaseModel):
        a: int
        b: list[str]

    features = PydanticToHFDatasets.model_cls_to_features(WithList)
    assert all([key in features for key in ("a", "b")])
    assert features["a"].dtype == "int32"
    assert isinstance(features["b"], datasets.Sequence)
    assert features["b"].feature.dtype == "string"
    kwargs = {"a": 1, "b": ["blub", "blab"]}
    dataset = datasets.Dataset.from_list([PydanticToHFDatasets.model_to_dict(WithList(**kwargs))], features=features)
    assert list(dataset)[0] == kwargs

    # With a dictionary of primitives.

    class WithDict(pydantic.BaseModel):
        a: int
        b: dict[str, int]
        c: dict

    features = PydanticToHFDatasets.model_cls_to_features(WithDict)
    assert all([key in features for key in ("a", "b")])
    assert isinstance(features["b"], datasets.Sequence)
    assert isinstance(features["b"].feature, datasets.Features)
    assert isinstance(features["c"], datasets.Value)
    assert features["c"].dtype == "string"
    assert all([name in features["b"].feature for name in ("key", "value")])
    dataset = datasets.Dataset.from_list(
        [PydanticToHFDatasets.model_to_dict(WithDict(a=1, b={"blub": 2, "blab": 3}, c={"blib": 4}))],
        features=features
    )
    assert list(dataset)[0] == {'a': 1, 'b': {'key': ['blub', 'blab'], 'value': [2, 3]}, 'c': "{'blib': 4}"}

    # With nested Pydantic models.

    class SubModel(pydantic.BaseModel):
        c: int

    class WithPydModel(pydantic.BaseModel):
        a: bool
        b: SubModel

    features = PydanticToHFDatasets.model_cls_to_features(WithPydModel)
    assert all([key in features for key in ("a", "b")])
    assert isinstance(features["a"], datasets.Value)
    assert features["a"].dtype == "bool"
    assert isinstance(features["b"], dict)
    assert "c" in features["b"]
    assert isinstance(features["b"]["c"], datasets.Value)
    assert features["b"]["c"].dtype == "int32"
    dataset = datasets.Dataset.from_list(
        [PydanticToHFDatasets.model_to_dict(WithPydModel(a=True, b=SubModel(c=3)))], features=features
    )
    assert list(dataset)[0] == {"a": True, "b": {"c": 3}}

    # With a dictionary of nested Pydantic models.

    class NestedModel(pydantic.BaseModel):
        sub_models: list[SubModel]

    class WithNestedDictPydModel(pydantic.BaseModel):
        a: bool
        b: dict[str, NestedModel]

    features = PydanticToHFDatasets.model_cls_to_features(WithNestedDictPydModel)
    assert all([key in features for key in ("a", "b")])
    assert isinstance(features["a"], datasets.Value)
    assert features["a"].dtype == "bool"
    assert isinstance(features["b"], datasets.Sequence)
    assert isinstance(features["b"].feature, datasets.Features)
    assert isinstance(features["b"].feature["key"], datasets.Value)
    assert features["b"].feature["key"].dtype == "string"
    assert isinstance(features["b"].feature["value"], datasets.Features)
    assert isinstance(features["b"].feature["value"]["sub_models"], datasets.Sequence)
    assert isinstance(features["b"].feature["value"]["sub_models"].feature, datasets.Features)
    assert isinstance(features["b"].feature["value"]["sub_models"].feature["c"], datasets.Value)
    assert features["b"].feature["value"]["sub_models"].feature["c"].dtype == "int32"


def test_pydantic_to_hf_nested_and_optional() -> None:
    """Test nested and optional models in PydanticToHFDatasets."""

    # Nested models in list
    class SubModel(pydantic.BaseModel):
        x: int

    class MainModelList(pydantic.BaseModel):
        subs: list[SubModel]

    model_list = MainModelList(subs=[SubModel(x=1), SubModel(x=2)])
    features_list = PydanticToHFDatasets.model_cls_to_features(MainModelList)
    result_list = PydanticToHFDatasets.model_to_dict(model_list)
    assert result_list == {"subs": [{"x": 1}, {"x": 2}]}
    ds_list = datasets.Dataset.from_list([result_list], features=features_list)
    assert ds_list[0] == {"subs": {"x": [1, 2]}}

    # Nested models in dict
    class MainModelDict(pydantic.BaseModel):
        subs: dict[str, SubModel]

    model_dict = MainModelDict(subs={"a": SubModel(x=1), "b": SubModel(x=2)})
    features_dict = PydanticToHFDatasets.model_cls_to_features(MainModelDict)
    result_dict = PydanticToHFDatasets.model_to_dict(model_dict)
    assert result_dict == {"subs": [{"key": "a", "value": {"x": 1}}, {"key": "b", "value": {"x": 2}}]}
    ds_dict = datasets.Dataset.from_list([result_dict], features=features_dict)
    assert ds_dict[0] == {"subs": {"key": ["a", "b"], "value": [{"x": 1}, {"x": 2}]}}

    # Optional nested model
    class MainModelOptional(pydantic.BaseModel):
        sub: SubModel | None

    model_opt = MainModelOptional(sub=SubModel(x=1))
    features_opt = PydanticToHFDatasets.model_cls_to_features(MainModelOptional)
    assert isinstance(features_opt["sub"], datasets.Features)
    result_opt = PydanticToHFDatasets.model_to_dict(model_opt)
    assert result_opt == {"sub": {"x": 1}}
    ds_opt = datasets.Dataset.from_list([result_opt], features=features_opt)
    assert ds_opt[0] == {"sub": {"x": 1}}

    # Deeply nested objects
    class Level3(pydantic.BaseModel):
        val: int

    class Level2(pydantic.BaseModel):
        l3: Level3

    class Level1(pydantic.BaseModel):
        l2: list[Level2]

    model_deep = Level1(l2=[Level2(l3=Level3(val=42))])
    features_deep = PydanticToHFDatasets.model_cls_to_features(Level1)
    result_deep = PydanticToHFDatasets.model_to_dict(model_deep)
    assert result_deep == {"l2": [{"l3": {"val": 42}}]}
    ds_deep = datasets.Dataset.from_list([result_deep], features=features_deep)
    assert ds_deep[0] == {"l2": {"l3": [{"val": 42}]}}



"""Tests for consolidation strategies."""

import pydantic
import pytest
from typing import Any

from sieves.tasks.predictive.consolidation import (
    MultiEntityConsolidation,
    SingleEntityConsolidation,
    LabelScoreConsolidation,
    TextConsolidation,
    QAConsolidation,
    MapScoreConsolidation,
)


class DummyEntity(pydantic.BaseModel):
    """Dummy entity for testing."""
    name: str
    score: float | None


def test_multi_entity_consolidation():
    strategy = MultiEntityConsolidation(extractor=lambda res: res["entities"])
    results = [
        {"entities": [DummyEntity(name="A", score=0.8), DummyEntity(name="B", score=0.6)]},
        {"entities": [DummyEntity(name="A", score=0.4)]},
    ]
    offsets = [(0, 2)]

    consolidated = strategy.consolidate(results, offsets)
    assert len(consolidated) == 1
    entities = sorted(consolidated[0], key=lambda x: x.name)
    assert len(entities) == 2
    assert entities[0].name == "A"
    assert entities[0].score == pytest.approx(0.6)
    assert entities[1].name == "B"
    assert entities[1].score == 0.6


def test_single_entity_consolidation():
    strategy = SingleEntityConsolidation(extractor=lambda res: res["entity"])
    results = [
        {"entity": DummyEntity(name="A", score=0.8)},
        {"entity": DummyEntity(name="A", score=0.6)},
        {"entity": DummyEntity(name="B", score=0.9)},
    ]
    offsets = [(0, 3)]

    consolidated = strategy.consolidate(results, offsets)
    assert len(consolidated) == 1
    winner = consolidated[0]
    assert winner is not None
    assert winner.name == "A"
    assert winner.score == pytest.approx(0.7)


def test_label_score_consolidation():
    labels = ["positive", "negative", "neutral"]
    strategy = LabelScoreConsolidation(
        labels=labels,
        mode="multi",
        extractor=lambda res: res
    )

    results = [
        {"positive": 0.8, "negative": 0.1, "neutral": 0.1},
        {"positive": 0.6, "negative": 0.2, "neutral": 0.2},
    ]
    offsets = [(0, 2)]

    consolidated = strategy.consolidate(results, offsets)
    assert len(consolidated) == 1
    scores = consolidated[0]
    # positive: (0.8 + 0.6) / 2 = 0.7
    # negative: (0.1 + 0.2) / 2 = 0.15
    # neutral: (0.1 + 0.2) / 2 = 0.15
    assert scores[0] == ("positive", pytest.approx(0.7))
    # negative and neutral might be in any order due to same score, but positive must be first
    assert scores[0][0] == "positive"

    # Test single mode
    strategy_single = LabelScoreConsolidation(
        labels=labels,
        mode="single",
        extractor=lambda res: {res["label"]: res["score"]}
    )
    results_single = [
        {"label": "positive", "score": 0.8},
        {"label": "negative", "score": 0.2},
    ]
    consolidated_single = strategy_single.consolidate(results_single, [(0, 2)])
    assert len(consolidated_single) == 1
    # positive: 0.8 / 2 = 0.4
    # negative: 0.2 / 2 = 0.1
    # neutral: 0.0
    assert consolidated_single[0][0] == ("positive", pytest.approx(0.4))


def test_text_consolidation():
    strategy = TextConsolidation(extractor=lambda res: (res["text"], res.get("score")), joiner=" ")
    results = [
        {"text": "Hello", "score": 0.8},
        {"text": "world", "score": 0.4},
    ]
    offsets = [(0, 2)]

    consolidated = strategy.consolidate(results, offsets)
    assert len(consolidated) == 1
    text, score = consolidated[0]
    assert text == "Hello world"
    assert score == pytest.approx(0.6)


def test_qa_consolidation():
    questions = ["What?", "Who?"]
    strategy = QAConsolidation(
        questions=questions,
        extractor=lambda res: [(qa["q"], qa["a"], qa.get("s")) for qa in res["qa"]]
    )

    results = [
        {"qa": [{"q": "What?", "a": "A", "s": 0.8}, {"q": "Who?", "a": "B", "s": 0.6}]},
        {"qa": [{"q": "What?", "a": "is", "s": 0.2}, {"q": "Who?", "a": "C", "s": None}]},
    ]
    offsets = [(0, 2)]

    consolidated = strategy.consolidate(results, offsets)
    assert len(consolidated) == 1
    qa_list = consolidated[0]
    assert len(qa_list) == 2

    # What? -> "A is", score (0.8+0.2)/2 = 0.5
    assert qa_list[0][0] == "What?"
    assert qa_list[0][1] == "A is"
    assert qa_list[0][2] == pytest.approx(0.5)

    # Who? -> "B C", score 0.6 / 1 = 0.6
    assert qa_list[1][0] == "Who?"
    assert qa_list[1][1] == "B C"
    assert qa_list[1][2] == pytest.approx(0.6)


def test_map_score_consolidation():
    aspects = ["quality", "price"]
    strategy = MapScoreConsolidation(
        keys=aspects,
        extractor=lambda res: (res["aspects"], res.get("score"))
    )

    results = [
        {"aspects": {"quality": 0.8, "price": 0.4}, "score": 0.9},
        {"aspects": {"quality": 0.6, "price": 0.2}, "score": 0.7},
    ]
    offsets = [(0, 2)]

    consolidated = strategy.consolidate(results, offsets)
    assert len(consolidated) == 1
    aspect_scores, overall_score = consolidated[0]

    assert aspect_scores["quality"] == pytest.approx(0.7)
    assert aspect_scores["price"] == pytest.approx(0.3)
    assert overall_score == pytest.approx(0.8)



"""Core task implementation."""

from __future__ import annotations

import abc
import itertools
import sys
from collections.abc import Callable, Iterable, Iterator
from typing import TYPE_CHECKING, Any

import dspy

from sieves.data import Doc
from sieves.serialization import Attribute, Config
from sieves.tasks.predictive.evaluation import TaskEvaluationReport

if TYPE_CHECKING:
    from sieves.pipeline import Pipeline


class Task(abc.ABC):
    """Abstract base class for tasks that can be executed on documents."""

    def __init__(
        self,
        task_id: str | None,
        include_meta: bool,
        batch_size: int,
        condition: Callable[[Doc], bool] | None = None,
    ):
        """
        Initiate new Task.

        :param task_id: Task ID.
        :param include_meta: Whether to include meta information generated by the task.
        :param batch_size: Batch size for processing documents. Use -1 to process all documents at once.
        :param condition: Optional callable that determines whether to process each document.
                          If provided, called with each Doc; if returns False, document is skipped
                          and results[task_id] is set to None.
        """
        self._task_id = task_id if task_id else self.__class__.__name__
        self._include_meta = include_meta
        self._batch_size = batch_size
        self._condition = condition

    @property
    def id(self) -> str:
        """Return task ID.

        Used by pipeline for results and dependency management.

        :return: Task ID.
        """
        return self._task_id

    def __call__(self, docs: Iterable[Doc]) -> Iterable[Doc]:
        """Execute task with conditional logic.

        Checks the condition for each document without materializing all docs upfront.
        Passes all documents that pass the condition to _call() for proper batching.
        Documents that fail the condition have results[task_id] set to None.

        :param docs: Docs to process.
        :return: Processed docs (in original order).
        """
        docs = iter(docs) if not isinstance(docs, Iterator) else docs

        # Materialize docs in batches. This doesn't incur additional memory overhead, as docs are materialized in
        # batches downstream anyway.
        batch_size = self._batch_size if self._batch_size > 0 else sys.maxsize
        while docs_batch := [doc for doc in itertools.islice(docs, batch_size)]:
            # First pass: determine which docs pass the condition by index.
            passing_indices: set[int] = {
                idx for idx, doc in enumerate(docs_batch) if self._condition is None or self._condition(doc)
            }

            # Process all passing docs in one batch.
            processed = self._call(d for i, d in enumerate(docs_batch) if i in passing_indices)
            processed_iter = iter(processed) if not isinstance(processed, Iterator) else processed

            # Iterate through original docs in order and yield results.
            for idx, doc in enumerate(docs_batch):
                if idx in passing_indices:
                    # Doc passed condition - use processed result.
                    yield next(processed_iter)
                else:
                    # Doc failed condition - set `None` result and yield original.
                    doc.results[self.id] = None
                    yield doc

    @abc.abstractmethod
    def _call(self, docs: Iterable[Doc]) -> Iterable[Doc]:
        """Execute task logic (to be implemented by subclasses).

        :param docs: Docs to process.
        :return: Processed docs.
        """

    def evaluate(self, docs: Iterable[Doc], judge: dspy.LM | None = None) -> TaskEvaluationReport:
        """Evaluate task performance.

        :param docs: Documents to evaluate.
        :param judge: Optional judge model for evaluation.
        :return: Evaluation report.
        """
        raise NotImplementedError(f"Evaluation not implemented for task {self.__class__.__name__}")

    def __add__(self, other: Task | Pipeline) -> Pipeline:
        """Chain this task with another task or pipeline using the ``+`` operator.

        This returns a new ``Pipeline`` that executes this task first, followed by the
        task(s) in ``other``. The original task(s)/pipeline are not mutated.

        Cache semantics:
        - If ``other`` is a ``Pipeline``, the resulting pipeline adopts ``other``'s
          ``use_cache`` setting (because the left-hand side is a single task).
        - If ``other`` is a ``Task``, the resulting pipeline defaults to ``use_cache=True``.

        :param other: A ``Task`` or ``Pipeline`` to execute after this task.
        :return: A new ``Pipeline`` representing the chained execution.
        :raises TypeError: If ``other`` is not a ``Task`` or ``Pipeline``.
        """
        # Lazy import to avoid circular dependency at module import time.
        from sieves.pipeline import Pipeline

        if isinstance(other, Pipeline):
            return Pipeline(tasks=[self, *other.tasks], use_cache=other.use_cache)

        if isinstance(other, Task):
            return Pipeline(tasks=[self, other])

        raise TypeError(f"Cannot chain Task with {type(other).__name__}")

    @property
    def _state(self) -> dict[str, Any]:
        """Return attributes to serialize.

        :return: Dict of attributes to serialize.
        """
        return {
            "task_id": self._task_id,
            "include_meta": self._include_meta,
            "batch_size": self._batch_size,
            "condition": self._condition,
        }

    def serialize(self) -> Config:
        """Serialize task.

        :return: Config instance.
        """
        return Config.create(self.__class__, {k: Attribute(value=v) for k, v in self._state.items()})

    @classmethod
    def deserialize(cls, config: Config, **kwargs: dict[str, Any]) -> Task:
        """Generate Task instance from config.

        :param config: Config to generate instance from.
        :param kwargs: Values to inject into loaded config.
        :return: Deserialized Task instance.
        """
        # Deserialize and inject model wrapper.
        return cls(**config.to_init_dict(cls, **kwargs))



"""
Shared fixtures for documentation example tests.

These fixtures provide commonly-used models and utilities for example tests,
loaded once per test session for efficiency.
"""

import pytest
import chonkie
import outlines
import tokenizers
import transformers
from transformers import AutoModelForCausalLM, AutoTokenizer


@pytest.fixture(scope="session")
def small_transformer_model():
    """
    Small HuggingFace transformer model for fast zero-shot classification.

    Uses MoritzLaurer/xtremedistil-l6-h256-zeroshot-v1.1-all-33 which is:
    - Fast (6 layers, 256 hidden size)
    - Accurate for zero-shot classification
    - No API key required
    """
    return transformers.pipeline(
        "zero-shot-classification",
        model="MoritzLaurer/xtremedistil-l6-h256-zeroshot-v1.1-all-33"
    )


@pytest.fixture(scope="session")
def example_tokenizer():
    """
    GPT-2 tokenizer for chunking examples.

    Used in chunking and preprocessing examples.
    """
    return tokenizers.Tokenizer.from_pretrained("gpt2")


@pytest.fixture(scope="session")
def example_chunker(example_tokenizer):
    """
    Chonkie TokenChunker for document chunking examples.

    Configured with:
    - chunk_size=512 tokens
    - chunk_overlap=50 tokens

    Uses GPT-2 tokenizer.
    """
    return chonkie.TokenChunker(
        example_tokenizer,
        chunk_size=512,
        chunk_overlap=50
    )


@pytest.fixture(scope="session")
def small_outlines_model():
    """
    Small Outlines model for information extraction and structured generation.

    Uses HuggingFaceTB/SmolLM-135M-Instruct which is:
    - Very small (135M parameters)
    - Fast inference
    - Supports structured generation
    - No API key required
    """
    model_name = "HuggingFaceTB/SmolLM-135M-Instruct"
    return outlines.models.from_transformers(
        AutoModelForCausalLM.from_pretrained(model_name),
        AutoTokenizer.from_pretrained(model_name)
    )


@pytest.fixture(scope="session")
def small_dspy_model():
    """
    Small DSPy model for optimization and distillation examples.

    Uses OpenRouter with a small, fast model.
    Requires OPENROUTER_API_KEY environment variable.
    """
    import os
    import dspy

    openrouter_api_base = "https://openrouter.ai/api/v1/"
    openrouter_model_id = "google/gemini-2.5-flash-lite-preview-09-2025"

    return dspy.LM(
        f"openrouter/{openrouter_model_id}",
        api_base=openrouter_api_base,
        api_key=os.environ['OPENROUTER_API_KEY'],
        cache=False
    )



"""Bridges for information extraction task."""

import abc
from collections.abc import Callable, Iterable, Sequence
from typing import Any, Literal, TypeVar, override

import dspy
import pydantic

from sieves.data import Doc
from sieves.model_wrappers import (
    ModelType,
    ModelWrapperInferenceMode,
    dspy_,
    langchain_,
    outlines_,
)
from sieves.model_wrappers.types import ModelSettings
from sieves.tasks.predictive.bridges import Bridge
from sieves.tasks.predictive.consolidation import (
    MultiEntityConsolidation,
    SingleEntityConsolidation,
)
from sieves.tasks.predictive.schemas.information_extraction import ResultMulti, ResultSingle

_BridgePromptSignature = TypeVar("_BridgePromptSignature")
_BridgeResult = TypeVar("_BridgeResult")


class InformationExtractionBridge(Bridge[_BridgePromptSignature, _BridgeResult, ModelWrapperInferenceMode], abc.ABC):
    """Abstract base class for information extraction bridges."""

    def __init__(
        self,
        task_id: str,
        prompt_instructions: str | None,
        entity_type: type[pydantic.BaseModel],
        model_settings: ModelSettings,
        mode: Literal["multi", "single"],
        prompt_signature: type[pydantic.BaseModel],
        model_type: ModelType,
        fewshot_examples: Sequence[pydantic.BaseModel] = (),
    ):
        """Initialize information extraction bridge.

        :param task_id: Task ID.
        :param prompt_instructions: Custom prompt instructions. If None, default instructions are used.
        :param entity_type: Object type to extract.
        :param model_settings: Settings for structured generation.
        :param mode: Extraction mode. If "multi", all occurrences of the entity are extracted. If "single", exactly one
            (or no) entity is extracted.
        :param prompt_signature: Unified Pydantic prompt signature.
        :param model_type: Model type.
        :param fewshot_examples: Few-shot examples.
        """
        assert model_type in {ModelType.dspy, ModelType.gliner, ModelType.langchain, ModelType.outlines}

        super().__init__(
            task_id=task_id,
            prompt_instructions=prompt_instructions,
            overwrite=False,
            model_settings=model_settings,
            prompt_signature=prompt_signature,
            model_type=model_type,
            fewshot_examples=fewshot_examples,
        )
        self._entity_type = entity_type
        self._mode = mode

        if self._mode == "multi":
            self._consolidation_strategy = MultiEntityConsolidation(extractor=self._get_multi_extractor())
        else:
            self._consolidation_strategy = SingleEntityConsolidation(extractor=self._get_single_extractor())

    @staticmethod
    def _get_multi_extractor() -> Callable[[Any], Iterable[pydantic.BaseModel]]:
        """Return a callable that extracts a list of entities from a raw chunk result.

        :return: Multi-extractor callable.
        """
        return lambda res: res.entities

    @staticmethod
    def _get_single_extractor() -> Callable[[Any], pydantic.BaseModel | None]:
        """Return a callable that extracts a single entity from a raw chunk result.

        :return: Single-extractor callable.
        """
        return lambda res: res.entity


class DSPyInformationExtraction(InformationExtractionBridge[dspy_.PromptSignature, dspy_.Result, dspy_.InferenceMode]):
    """DSPy bridge for information extraction."""

    @override
    def _validate(self) -> None:
        assert self._model_type == ModelType.dspy

    @override
    @property
    def _default_prompt_instructions(self) -> str:
        return ""

    @override
    @property
    def inference_mode(self) -> dspy_.InferenceMode:
        return self._model_settings.inference_mode or dspy_.InferenceMode.predict

    @override
    def integrate(self, results: Sequence[dspy_.Result], docs: list[Doc]) -> list[Doc]:
        for doc, result in zip(docs, results):
            if self._mode == "multi":
                assert len(result.completions.entities) == 1
                doc.results[self._task_id] = ResultMulti(
                    entities=result.completions.entities[0],
                )
            else:
                assert len(result.completions.entity) == 1
                doc.results[self._task_id] = ResultSingle(
                    entity=result.completions.entity[0],
                )
        return docs

    @override
    def consolidate(
        self, results: Sequence[dspy_.Result], docs_offsets: list[tuple[int, int]]
    ) -> Sequence[dspy_.Result]:
        consolidated_results_clean = self._consolidation_strategy.consolidate(results, docs_offsets)

        # Wrap back into dspy.Prediction.
        consolidated_results: list[dspy_.Result] = []
        for res_clean in consolidated_results_clean:
            if self._mode == "multi":
                consolidated_results.append(
                    dspy.Prediction.from_completions(
                        {"entities": [res_clean]},
                        signature=self.prompt_signature,
                    )
                )
            else:
                consolidated_results.append(
                    dspy.Prediction.from_completions(
                        {"entity": [res_clean]},
                        signature=self.prompt_signature,
                    )
                )

        return consolidated_results


class PydanticInformationExtraction(
    InformationExtractionBridge[pydantic.BaseModel, pydantic.BaseModel, ModelWrapperInferenceMode],
):
    """Base class for Pydantic-based information extraction bridges."""

    @override
    def _validate(self) -> None:
        assert self._model_type in {ModelType.langchain, ModelType.outlines}

    @override
    @property
    def _default_prompt_instructions(self) -> str:
        if self._mode == "multi":
            return (
                "Find all occurences of this kind of entitity within the text. For each entity found, also provide "
                "a confidence score between 0.0 and 1.0 in the 'score' field."
            )

        return (
            "Find the single most relevant entitity within the text. If no such entitity exists, return null. "
            "Return exactly one entity with all its fields, NOT just a string. Also provide a confidence score "
            "between 0.0 and 1.0 in the 'score' field."
        )

    @override
    @property
    def _prompt_conclusion(self) -> str | None:
        return "========\n\n<text>{{ text }}</text>"

    @override
    def integrate(self, results: Sequence[pydantic.BaseModel], docs: list[Doc]) -> list[Doc]:
        for doc, result in zip(docs, results):
            if self._mode == "multi":
                assert hasattr(result, "entities")
                doc.results[self._task_id] = ResultMulti(entities=result.entities)
            else:
                assert hasattr(result, "entity")
                doc.results[self._task_id] = ResultSingle(entity=result.entity)
        return docs

    @override
    def consolidate(
        self,
        results: Sequence[pydantic.BaseModel],
        docs_offsets: list[tuple[int, int]],
    ) -> Sequence[pydantic.BaseModel]:
        assert issubclass(self.prompt_signature, pydantic.BaseModel)

        consolidated_results_clean = self._consolidation_strategy.consolidate(results, docs_offsets)
        consolidated_results: list[pydantic.BaseModel] = []

        for res_clean in consolidated_results_clean:
            if self._mode == "multi":
                consolidated_results.append(self.prompt_signature(entities=res_clean))
            else:
                consolidated_results.append(self.prompt_signature(entity=res_clean))

        return consolidated_results

    @override
    @property
    def inference_mode(self) -> outlines_.InferenceMode | langchain_.InferenceMode:
        if self._model_type == ModelType.outlines:
            return self._model_settings.inference_mode or outlines_.InferenceMode.json

        elif self._model_type == ModelType.langchain:
            return self._model_settings.inference_mode or langchain_.InferenceMode.structured

        raise ValueError(f"Unsupported model type: {self._model_type}")



# mypy: ignore-errors
import pytest
from sieves import Doc, Pipeline
from sieves.model_wrappers import ModelType
from sieves.tasks import Classification

@pytest.mark.parametrize("include_meta", [True, False])
@pytest.mark.parametrize("runtime", [ModelType.gliner], indirect=["runtime"])
def test_include_meta_toggle(classification_docs, include_meta, runtime):
    """Test whether the include_meta flag correctly toggles raw result population."""
    docs = [Doc(text=d.text) for d in classification_docs]
    task_id = "toggle_task"

    task = Classification(
        labels=["science", "politics"],
        task_id=task_id,
        model=runtime.model,
        model_settings=runtime.model_settings,
        include_meta=include_meta,
        batch_size=runtime.batch_size,
    )

    pipe = Pipeline(task)
    results = list(pipe(docs))

    for doc in results:
        print(f"Output: {doc.results[task_id]}")
        if include_meta:
            assert task_id in doc.meta
            assert "raw" in doc.meta[task_id]
            assert isinstance(doc.meta[task_id]["raw"], list)
            assert len(doc.meta[task_id]["raw"]) > 0
            print(f"Raw output: {doc.meta[task_id]['raw']}")
            print(f"Task Usage: {doc.meta[task_id]['usage']}")
            print(f"Total Usage: {doc.meta['usage']}")
        else:
            assert task_id not in doc.meta



"""Information extraction."""

from __future__ import annotations

import types
import typing
from collections.abc import Callable, Iterable, Sequence
from pathlib import Path
from typing import Any, Literal, override

import datasets
import dspy
import pydantic

from sieves.data import Doc
from sieves.model_wrappers import ModelType, gliner_
from sieves.model_wrappers.types import ModelSettings
from sieves.serialization import Config
from sieves.tasks.distillation.types import DistillationFramework
from sieves.tasks.predictive.core import PredictiveTask
from sieves.tasks.predictive.gliner_bridge import GliNERBridge
from sieves.tasks.predictive.information_extraction.bridges import (
    DSPyInformationExtraction,
    PydanticInformationExtraction,
)
from sieves.tasks.predictive.schemas.information_extraction import (
    FewshotExampleMulti,
    FewshotExampleSingle,
    TaskModel,
    TaskPromptSignature,
    TaskResult,
)
from sieves.tasks.utils import PydanticToHFDatasets

_TaskBridge = GliNERBridge | DSPyInformationExtraction | PydanticInformationExtraction

FewshotExample = FewshotExampleMulti | FewshotExampleSingle


def _wrap_with_score(entity_type: type[pydantic.BaseModel]) -> type[pydantic.BaseModel]:
    """Create a subclass of the provided entity type that includes a score field.

    :param entity_type: The Pydantic model class to wrap.
    :return: A new Pydantic model class with a score field.
    """
    if hasattr(entity_type, "score"):
        return entity_type

    class ScoredEntity(entity_type):  # type: ignore[valid-type, misc]
        """Extracted entity with an additional confidence score."""

        score: float | None = pydantic.Field(
            default=None, description="Provide a confidence score for the entity extraction, between 0 and 1."
        )

    ScoredEntity.__name__ = entity_type.__name__

    return ScoredEntity


class InformationExtraction(PredictiveTask[TaskPromptSignature, TaskResult, _TaskBridge]):
    """Information extraction task."""

    def __init__(
        self,
        entity_type: type[pydantic.BaseModel],
        model: TaskModel,
        task_id: str | None = None,
        include_meta: bool = True,
        batch_size: int = -1,
        prompt_instructions: str | None = None,
        fewshot_examples: Sequence[FewshotExample] = (),
        mode: Literal["multi", "single"] = "multi",
        model_settings: ModelSettings = ModelSettings(),
        condition: Callable[[Doc], bool] | None = None,
    ) -> None:
        """Initialize new PredictiveTask.

        :param entity_type: Pydantic model class representing the object type to extract.
        :param model: Model to use.
        :param task_id: Task ID.
        :param include_meta: Whether to include meta information generated by the task.
        :param batch_size: Batch size to use for inference. Use -1 to process all documents at once.
        :param prompt_instructions: Custom prompt instructions. If None, default instructions are used.
        :param fewshot_examples: Few-shot examples.
        :param mode: Extraction mode. If "multi", all occurrences of the entity are extracted. If "single", exactly one
            (or no) entity is extracted.
        :param model_settings: Settings for structured generation.
        :param condition: Optional callable that determines whether to process each document.
        """
        self._entity_type = entity_type
        self._mode = mode
        self._scored_entity_type = _wrap_with_score(entity_type)

        super().__init__(
            model=model,
            task_id=task_id,
            include_meta=include_meta,
            batch_size=batch_size,
            overwrite=False,
            prompt_instructions=prompt_instructions,
            fewshot_examples=fewshot_examples,
            model_settings=model_settings,
            condition=condition,
        )

        if (
            isinstance(entity_type, type)
            and issubclass(entity_type, pydantic.BaseModel)
            and not entity_type.model_config.get("frozen", False)
        ):
            raise ValueError(
                f"Entity type provided to task {self._task_id} isn't frozen, which means that entities can't "
                f"be deduplicated and compared. Modify entity_type to be frozen=True."
            )

    @property
    @override
    def fewshot_example_type(self) -> type[FewshotExample]:
        """Return few-shot example type.

        :return: Few-shot example type.
        """
        if self._mode == "multi":
            return FewshotExampleMulti
        return FewshotExampleSingle

    @property
    @override
    def prompt_signature(self) -> type[pydantic.BaseModel]:
        """Return the unified Pydantic prompt signature for this task.

        :return: Unified Pydantic prompt signature.
        """
        scored_type = self._scored_entity_type

        if self._mode == "multi":
            return pydantic.create_model(
                f"{scored_type.__name__}Multi",
                __doc__=f"Result of multi-entity extraction for {scored_type.__name__}.",
                entities=(
                    list[scored_type],  # type: ignore[valid-type]
                    pydantic.Field(..., description=f"List of extracted {scored_type.__name__} entities."),
                ),
            )
        return pydantic.create_model(
            f"{scored_type.__name__}Single",
            __doc__=f"Result of single-entity extraction for {scored_type.__name__}.",
            entity=(
                scored_type | None,  # type: ignore[valid-type]
                pydantic.Field(..., description=f"The extracted {scored_type.__name__} entity."),
            ),
        )

    @property
    @override
    def metric(self) -> str:
        return "F1" if self._mode == "multi" else "Accuracy"

    @override
    def _compute_metrics(self, truths: list[Any], preds: list[Any], judge: dspy.LM | None = None) -> dict[str, float]:
        """Compute corpus-level metrics.

        :param truths: List of ground truths.
        :param preds: List of predictions.
        :param judge: Optional DSPy LM instance to use as judge for generative tasks.
        :return: Dictionary of metrics.
        """
        # Single mode.
        if self._mode == "single":
            correct = 0
            total = 0
            for gold, pred in zip(truths, preds):
                # If gold is None, we assume it's a negative example (no entity).
                # If pred is None, it predicted no entity.
                # If both None -> Correct.
                if gold is not None:
                    assert isinstance(gold, TaskResult)
                    gold_entity = gold.entity
                else:
                    gold_entity = None

                if pred is not None:
                    assert isinstance(pred, TaskResult)
                    pred_entity = pred.entity
                else:
                    pred_entity = None

                if gold_entity == pred_entity:
                    correct += 1
                total += 1

            return {self.metric: correct / total if total > 0 else 0.0}

        # Multi mode.
        tp = 0
        fp = 0
        fn = 0

        def entity_to_tuple(entity: Any) -> tuple:
            # Handle Pydantic model
            d = entity.model_dump()

            # Remove score if present
            if "score" in d:
                del d["score"]

            items = sorted(d.items())
            return tuple((k, v if not (isinstance(v, list) or isinstance(v, dict)) else str(v)) for k, v in items)

        for gold, pred in zip(truths, preds):
            if gold is not None:
                assert isinstance(gold, TaskResult)
                true_entities = {entity_to_tuple(e) for e in gold.entities}
            else:
                true_entities = set()

            if pred is not None:
                assert isinstance(pred, TaskResult)
                pred_entities = {entity_to_tuple(e) for e in pred.entities}
            else:
                pred_entities = set()

            tp += len(true_entities & pred_entities)
            fp += len(pred_entities - true_entities)
            fn += len(true_entities - pred_entities)

        precision = tp / (tp + fp) if (tp + fp) > 0 else 0.0
        recall = tp / (tp + fn) if (tp + fn) > 0 else 0.0
        f1 = 2 * precision * recall / (precision + recall) if (precision + recall) > 0 else 0.0

        return {self.metric: f1}

    @override
    def _init_bridge(self, model_type: ModelType) -> _TaskBridge:
        """Initialize bridge.

        :param model_type: Type of model to initialize bridge for.
        :return _TaskBridge: ModelWrapper task bridge.
        :raises ValueError: If model type is not supported.
        :raises TypeError: On entity type and model type mismatch.
        """
        if model_type == ModelType.gliner:
            return GliNERBridge(
                task_id=self._task_id,
                prompt_instructions=self._custom_prompt_instructions,
                prompt_signature=self.prompt_signature,
                model_settings=self._model_settings,
                inference_mode=gliner_.InferenceMode.structure,
                mode=self._mode,
            )

        bridge_types: dict[ModelType, type[_TaskBridge]] = {
            ModelType.dspy: DSPyInformationExtraction,
            ModelType.langchain: PydanticInformationExtraction,
            ModelType.outlines: PydanticInformationExtraction,
        }

        try:
            bridge = bridge_types[model_type](
                task_id=self._task_id,
                prompt_instructions=self._custom_prompt_instructions,
                entity_type=self._scored_entity_type,
                model_settings=self._model_settings,
                mode=self._mode,
                prompt_signature=self.prompt_signature,
                model_type=model_type,
                fewshot_examples=self._fewshot_examples,
            )

        except KeyError as err:
            raise KeyError(f"Model type {model_type} is not supported by {self.__class__.__name__}.") from err

        return bridge

    @staticmethod
    @override
    def supports() -> set[ModelType]:
        return {
            ModelType.dspy,
            ModelType.gliner,
            ModelType.langchain,
            ModelType.outlines,
        }

    @override
    @property
    def _state(self) -> dict[str, Any]:
        return {
            **super()._state,
            "entity_type": self._entity_type,
            "mode": self._mode,
        }

    @override
    def _validate_fewshot_examples(self) -> None:
        example_type_error_text = "Fewshot example type mismatch: mode = {mode} requires {example_type}."

        for i, fs_example in enumerate(self._fewshot_examples or []):
            if self._mode == "multi":
                assert isinstance(fs_example, FewshotExampleMulti), TypeError(
                    example_type_error_text.format(example_type=FewshotExampleMulti, mode=self._mode)
                )
            else:
                assert isinstance(fs_example, FewshotExampleSingle), TypeError(
                    example_type_error_text.format(example_type=FewshotExampleSingle, mode=self._mode)
                )

    @override
    def to_hf_dataset(self, docs: Iterable[Doc], threshold: float | None = None) -> datasets.Dataset:
        # Use the scored entity type for the dataset schema.
        # We need to unwrap the unified signature container if it exists.
        entity_type = self.prompt_signature
        if "entities" in entity_type.model_fields:
            entity_type = entity_type.model_fields["entities"].annotation
            if typing.get_origin(entity_type) is list:
                entity_type = typing.get_args(entity_type)[0]
        elif "entity" in entity_type.model_fields:
            entity_type = entity_type.model_fields["entity"].annotation
            if typing.get_origin(entity_type) in (typing.Union, types.UnionType):
                args = typing.get_args(entity_type)
                entity_type = [t for t in args if t is not type(None)][0]

        # Define metadata.
        hf_entity_type = PydanticToHFDatasets.model_cls_to_features(entity_type)
        if self._mode == "multi":
            features = datasets.Features(
                {"text": datasets.Value("string"), "entities": datasets.Sequence(hf_entity_type)}
            )
        else:
            features = datasets.Features({"text": datasets.Value("string"), "entity": hf_entity_type})

        entity_type_name = getattr(entity_type, "__name__", str(entity_type))
        info = datasets.DatasetInfo(
            description=f"Information extraction dataset for entity type {entity_type_name}. "
            f"Generated with sieves v{Config.get_version()}.",
            features=features,
        )

        # Fetch data used for generating dataset.
        try:
            if self._mode == "multi":
                data = [
                    (doc.text, [PydanticToHFDatasets.model_to_dict(res) for res in doc.results[self._task_id].entities])
                    for doc in docs
                ]
            else:
                data = [
                    (
                        doc.text,
                        PydanticToHFDatasets.model_to_dict(doc.results[self._task_id].entity)
                        if doc.results[self._task_id].entity
                        else None,
                    )
                    for doc in docs
                ]
        except KeyError as err:
            raise KeyError(f"Not all documents have results for this task with ID {self._task_id}") from err

        # Create dataset.
        return datasets.Dataset.from_list(
            [
                {
                    "text": text,
                    ("entities" if self._mode == "multi" else "entity"): res,
                }
                for text, res in data
            ],
            features=features,
            info=info,
        )

    @override
    def distill(
        self,
        base_model_id: str,
        framework: DistillationFramework,
        data: datasets.Dataset | Sequence[Doc],
        output_path: Path | str,
        val_frac: float,
        init_kwargs: dict[str, Any] | None = None,
        train_kwargs: dict[str, Any] | None = None,
        seed: int | None = None,
    ) -> None:
        raise NotImplementedError

    @override
    def _evaluate_dspy_example(self, truth: dspy.Example, pred: dspy.Prediction, trace: Any, model: dspy.LM) -> float:
        if self._mode == "single":
            true_entity = truth.get("entity")
            pred_entity = pred.get("entity")
            base_accuracy = 1.0 if true_entity == pred_entity else 0.0

            # If score is available in both truth and pred, incorporate it into the metric.
            if "score" in truth and truth["score"] is not None and "score" in pred and pred["score"] is not None:
                score_accuracy = 1 - abs(truth["score"] - max(min(pred["score"], 1), 0))
                return (base_accuracy + score_accuracy) / 2

            return base_accuracy

        def entity_to_tuple(entity: dict) -> tuple:
            """Convert entity dict to hashable tuple for comparison.

            Converts nested structures (lists, dicts) to strings to make them hashable.

            :param entity: Entity dictionary to convert.
            :return: Hashable tuple representation of the entity.
            """
            items = sorted(entity.items())
            return tuple((k, v if not (isinstance(v, list) or isinstance(v, dict)) else str(v)) for k, v in items)

        # Compute set-based F1 score for entity extraction
        true_entities = {entity_to_tuple(e) for e in truth["entities"]}
        pred_entities = {entity_to_tuple(e) for e in pred.get("entities", [])}

        if not true_entities:
            return 1.0 if not pred_entities else 0.0

        precision = len(true_entities & pred_entities) / len(pred_entities) if pred_entities else 0
        recall = len(true_entities & pred_entities) / len(true_entities)
        f1 = 2 * precision * recall / (precision + recall) if (precision + recall) > 0 else 0

        # If scores are available, incorporate them.
        if "scores" in truth and truth["scores"] is not None and "scores" in pred and pred["scores"] is not None:
            # Simple averaging of score similarities for matched entities is complex here due to set-based matching.
            # We'll just average the overall entity f1 with a score similarity metric.
            true_avg_score = sum(truth["scores"]) / len(truth["scores"]) if truth["scores"] else 0
            pred_avg_score = sum(pred["scores"]) / len(pred["scores"]) if pred["scores"] else 0
            score_accuracy = 1 - abs(true_avg_score - max(min(pred_avg_score, 1), 0))
            return (f1 + score_accuracy) / 2

        return f1
